#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
import pandas as pd

from cdflib import cdfepoch, cdfread
from dateutil import parser as date_parser

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _datetime_to_tt2000(time):
    r"""Converts datetime.datetime into epoch_tt_2000.

    Parameters
    ----------
    time : datetime.datetime or list of datetime.datetime
        Time or list of times to convert.

    Returns
    -------
    tt2000 : str
        Time into TT2000 format.

    """

    time_ = pd.Timestamp(time)

    # Convert to string
    return f"{time_.strftime('%Y-%m-%dT%H:%M:%S.%f')}{time_.nanosecond:03d}"


def _rcdf(path, tint):
    r"""Reads CDF files.

    Parameters
    ----------
    path : str
        String of the filename in .cdf containing the L2 data
    tint : list
        Time interval

    Returns
    -------
    out_dict : dict
        Hash table with fields contained in the .cdf file.

    """
    out_dict = {}

    with cdfread.CDF(path) as file:
        keys_ = file.cdf_info()["zVariables"]
        for k_ in keys_:
            temp_ = file.varget(k_, starttime=tint[0], endtime=tint[1])
            shape_ = temp_.shape
            coords = [np.arange(lim_) for lim_ in shape_]

            out_dict[k_.lower()] = xr.DataArray(temp_, coords=coords)

    return out_dict


def read_tnr(path, tint, sensor: int = 4):
    r"""Read L2 data from TNR

    Parameters
    ----------
    path : str
        String of the filename in .cdf containing the L2 data
    tint : list
        Time interval
    sensor : int, Optional
        TNR sensor to be read:
            * 1: V1
            * 2: V2
            * 3: V3,
            * 4: V1 - V2
            * 5: V2 - V3
            * 6: V3 - V1
            * 7: B

    Returns
    -------
    out : xarray.DataArray
        Spectrum of the measured signals.

    Notes
    -----
    The script check if there are data from the two channel and put them
    together.

    """

    tint_ = list(map(date_parser.parse, tint))
    tint_ = list(map(_datetime_to_tt2000, tint_))
    tint_ = list(map(cdfepoch.parse, tint_))

    data_l2 = _rcdf(path, tint_)

    n_freqs = 4 * data_l2["tnr_band_freq"].shape[1]
    freq_tnr = np.reshape(data_l2["tnr_band_freq"].data, n_freqs) / 1000

    epoch_ = data_l2["epoch"].data
    auto1_ = data_l2["auto1"].data
    auto2_ = data_l2["auto2"].data
    sweep_ = data_l2["sweep_num"].data
    bande_ = data_l2["tnr_band"].data
    confg_ = data_l2["sensor_config"].data

    if sensor == 7:
        auto1_ = data_l2["magnetic_spectral_power1"].data
        auto2_ = data_l2["magnetic_spectral_power2"].data

    puntical = np.where(data_l2["front_end"].data == 1)[0]

    epoch_ = epoch_[puntical]
    confg_ = confg_[puntical, :]
    auto1_ = auto1_[puntical, :]
    auto2_ = auto2_[puntical, :]
    sweep_ = sweep_[puntical]
    bande_ = bande_[puntical]

    sweep_num = sweep_

    delta_sw = np.abs(sweep_[1:] - sweep_[:-1])

    xdelta_sw = np.where(delta_sw > 100)[0]
    if xdelta_sw.size:
        xdelta_sw = np.hstack([xdelta_sw, len(sweep_) - 1])
        nxdelta_sw = len(xdelta_sw)
        for inswn in range(nxdelta_sw - 1):
            idx_l, idx_r = [xdelta_sw[inswn] + 1, xdelta_sw[inswn + 1]]
            sweep_num[idx_l:idx_r] += sweep_num[xdelta_sw[inswn]]

    timet_ = cdfepoch.to_datetime(epoch_, to_np=True)

    sens0_, sens1_ = [np.where(confg_[:, i] == sensor)[0] for i in range(2)]

    if sens0_.size and sens1_.size:
        auto_calib = np.vstack([auto1_[sens0_, :], auto2_[sens1_, :]])
        sens_ = np.hstack([sens0_, sens1_])
        timet_ici = np.hstack([timet_[sens0_], timet_[sens1_]])
    elif sens0_.size:
        auto_calib = auto1_[sens0_, :]
        sens_ = sens0_
        timet_ici = timet_[sens0_]
    elif sens1_.size:
        auto_calib = auto2_[sens1_, :]
        sens_ = sens1_
        timet_ici = timet_[sens1_]

    else:
        raise ValueError("no data at all ?!?")

    ord_time = np.argsort(timet_ici)
    time_rr = timet_ici[ord_time]
    sens_ = sens_[ord_time]
    auto_calib = auto_calib[ord_time, :]

    bande_e = bande_[sens_]
    max_sweep = np.max(sweep_num[sens_])
    min_sweep = np.min(sweep_num[sens_])
    sweep_num = sweep_num[sens_]

    v_, time, sweepn_tnr = [[], [], []]

    for ind_sweep in range(min_sweep, max_sweep):
        v1_ = np.zeros(128)
        p_punt = np.where(sweep_num == ind_sweep)[0]
        if p_punt.size:
            for indband in range(p_punt.size):
                idx_l = 32 * bande_e[p_punt[indband]]
                idx_r = 32 * (bande_e[p_punt[indband]] + 1)
                v1_[idx_l:idx_r] = auto_calib[p_punt[indband], :]

            if np.sum(v1_) > 0.0:
                punt0_ = np.where(v1_ == 0.0)[0]
                if punt0_.size:
                    v1_[punt0_] = np.nan
                v_.append(v1_)
                sweepn_tnr.append(sweep_num[p_punt[0]])

        if p_punt.size:
            time.append(time_rr[np.min(p_punt)])

    out = xr.DataArray(np.stack(v_), coords=[np.stack(time), freq_tnr])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import string

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def make_labels(axs, pos, pad: float = 0):
    r"""Add subplots labels to axes

    Parameters
    ----------
    axs : ndarray
        Array of subplots axes.
    pos : array_like
        Position of the text in the axis.
    pad : int, Optional
        Offset in axis counter.

    Returns
    -------
    axs : ndarray
        Array of subplots axes with labels.

    """

    lbl = string.ascii_lowercase[pad:len(axs) + pad]

    for label, axis in zip(lbl, axs):
        if "proj" in axis.properties():
            axis.text2D(pos[0], pos[1], "({})".format(label),
                        transform=axis.transAxes,
                        bbox=dict(boxstyle="square", ec=(1., 1., 1.),
                                  fc=(1., 1., 1.)))
        else:
            axis.text(pos[0], pos[1], "({})".format(label),
                      transform=axis.transAxes,
                      bbox=dict(boxstyle="square", ec=(1., 1., 1.),
                                fc=(1., 1., 1.)))

    return axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def vdf_omni(vdf):
    r"""Computes omni-directional distribution, without changing the units.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the 3D velocity distribution with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    Returns
    -------
    out : xarray.DataArray
        Time series of the omnidirectional velocity distribution function.

    """

    time = vdf.time.data

    energy = vdf.energy.data
    thetas = vdf.theta.data
    dangle = np.pi / vdf.theta.shape[0]
    np_phi = vdf.phi.shape[1]

    sine_theta = np.ones((np_phi, 1)) * np.sin(np.deg2rad(thetas))
    solid_angles = dangle * dangle * sine_theta
    all_solid_angles = np.tile(solid_angles,
                               (len(time), energy.shape[1], 1, 1))

    dist = vdf.data.data * all_solid_angles
    omni = np.squeeze(np.nanmean(np.nanmean(dist, axis=3), axis=2))
    omni /= np.mean(np.mean(solid_angles))

    energy = np.mean(energy[:2, :], axis=0)

    out = xr.DataArray(omni, coords=[time, energy], dims=["time", "energy"],
                       attrs=vdf.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eis_spin_avg(eis_allt, method: str = "mean"):
    r"""Calculates spin-averaged fluxes for the EIS instrument.

    Parameters
    ----------
    eis_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.
    method : {"mean", "sum"}
        Method.
    Returns
    -------
    spin_avg_fluxes : xarray.Dataset
        Spin-averaged fluxes for all 6 telescopes.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_omni

    Examples
    --------
    >>> from pyrfu import mms

    Define spacecraft index and time interval

    >>> tint = ["2017-07-23T16:10:00", "2017-07-23T18:10:00"]
    >>> ic = 2

    Get EIS ExTOF all 6 telescopes fluxes

    >>> extof_allt = mms.get_eis_allt("flux_extof_proton_srvy_l2", tint, ic)

    Spin average all 6 telescopes fluxes

    >>> extof_allt_despin = mms.eis_spin_avg(extof_allt)

    """

    spin_nums = eis_allt.spin

    spin_starts = np.where(spin_nums.data[1:] > spin_nums.data[:-1])[0]

    scopes = list(filter(lambda x: x[0] == "t", eis_allt))

    spin_avg_flux, spin_sum_flux = [{}, {}]

    for scope in scopes:
        scope_data = eis_allt[scope]

        time_recs = scope_data.time.data[spin_starts]
        energies_ = scope_data.energy.data
        flux_data = scope_data.data

        flux_av = np.zeros([len(spin_starts), len(energies_)])
        flux_su = np.zeros([len(spin_starts), len(energies_)])

        c_strt = 0

        for i, spin_strt in enumerate(spin_starts):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                flux_av[i, :] = np.nanmean(flux_data[c_strt:spin_strt + 1, :],
                                           axis=0)
                flux_su[i, :] = np.nansum(flux_data[c_strt:spin_strt + 1,
                                          :], axis=0)
            c_strt = spin_strt + 1

        spin_avg_flux[scope] = xr.DataArray(flux_av,
                                            coords=[time_recs, energies_],
                                            dims=["time", "energy"])
        spin_sum_flux[scope] = xr.DataArray(flux_su,
                                            coords=[time_recs, energies_],
                                            dims=["time", "energy"])

    spin_avg_flux["spin"] = eis_allt["spin"][spin_starts]
    spin_avg_flux["energy_dplus"] = eis_allt.energy_dplus.data
    spin_avg_flux["energy_dminus"] = eis_allt.energy_dminus.data
    spin_avg_flux = xr.Dataset(spin_avg_flux, attrs=eis_allt.attrs)

    spin_sum_flux["spin"] = eis_allt["spin"][spin_starts]
    spin_sum_flux["energy_dplus"] = eis_allt.energy_dplus.data
    spin_sum_flux["energy_dminus"] = eis_allt.energy_dminus.data
    spin_sum_flux = xr.Dataset(spin_sum_flux, attrs=eis_allt.attrs)

    out = [spin_avg_flux, spin_sum_flux][method.lower() != "mean"]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""one_fluid_dispersion.py
@author: Louis Richard
"""

import numpy as np
import xarray as xr

from scipy import constants, optimize


def _disprel(w, *args):
    k, theta = args[0:2]
    v_a, c_s = args[2:4]
    wc_e, wc_p = args[4:6]

    theta = np.deg2rad(theta)
    l_00 = 1
    l_01 = - w ** 2 / (k ** 2 * v_a ** 2)
    l_02 = - w ** 2 / (wc_e * wc_p)
    l_03 = k ** 2 * np.sin(theta) ** 2 / ((w / c_s) ** 2 - k ** 2)
    l_0_ = (l_00 + l_01 + l_02 + l_03)

    l_10 = np.cos(theta) ** 2
    l_11 = - w ** 2 / (k ** 2 * v_a ** 2)
    l_12 = - w ** 2 / (wc_e * wc_p)
    l_1_ = l_10 + l_11 + l_12

    r_0_ = w ** 2 * np.cos(theta) ** 2 / wc_p ** 2

    disprel = l_0_ * l_1_ - r_0_

    return disprel


def one_fluid_dispersion(b_0, theta, ions, electrons, n_k: int = 100):
    r"""Solves the one fluid dispersion relation.

    Parameters
    ----------
    b_0 : float
        Magnetic field

    theta : float
        The angle of propagation of the wave with respect to the magnetic
        field, :math:`\cos^{-1}(k_z / k)`

    ions : dict
        Hash table with n : number density, t: temperature, gamma:
        polytropic index.

    electrons : dict
        Hash table with n : number density, t: temperature, gamma:
        polytropic index.

    n_k : int, optional
        Number of wavenumbers.

    Returns
    -------
    wc_1 : xarray.DataArray
        1st root

    wc_2 : xarray.DataArray
        2nd root

    wc_3 : xarray.DataArray
        3rd root

    """

    keys = ["n", "t", "gamma"]
    n_p, t_p, gamma_p = [ions[k] for k in keys]
    n_e, t_e, gamma_e = [electrons[k] for k in keys]

    q_e = constants.elementary_charge
    m_e = constants.electron_mass
    m_p = constants.proton_mass
    ep_0 = constants.epsilon_0
    mu_0 = constants.mu_0

    wc_e = q_e * b_0 / m_e
    wc_p = q_e * b_0 / m_p

    wp_e = np.sqrt(q_e ** 2 * n_e / (ep_0 * m_e))
    wp_p = np.sqrt(q_e ** 2 * n_p / (ep_0 * m_p))

    v_p = np.sqrt(q_e * t_p / m_p)
    v_e = np.sqrt(q_e * t_e / m_e)

    v_a = b_0 / np.sqrt(mu_0 * n_p * m_p)
    c_s = np.sqrt((gamma_e * q_e * t_e + gamma_p * q_e * t_p) / (m_e + m_p))

    k_vec = np.linspace(2e-7, 1.0e-4, n_k)

    wc_1, wc_2, wc_3 = [np.zeros(len(k_vec)) for _ in range(3)]

    for i, k in enumerate(k_vec):
        if i < 10:
            guess_w1 = v_a * k * 1.50
            guess_w2 = v_a * k * 0.70
            guess_w3 = c_s * k * 0.99
        else:
            guess_w1 = wc_1[i - 1] + (wc_1[i - 1] - wc_1[i - 2])
            guess_w2 = wc_2[i - 1] + (wc_2[i - 1] - wc_2[i - 2])
            guess_w3 = wc_3[i - 1] + (wc_3[i - 1] - wc_3[i - 2])

        arguments = (k, theta, v_a, c_s, wc_e, wc_p)
        wc_1[i] = optimize.fsolve(_disprel, guess_w1, args=arguments)[0]
        wc_2[i] = optimize.fsolve(_disprel, guess_w2, args=arguments)[0]
        wc_3[i] = optimize.fsolve(_disprel, guess_w3, args=arguments)[0]

    attrs = {"wc_e": wc_e, "wc_p": wc_p, "wp_e": wp_e, "wp_p": wp_p,
             "v_p": v_p, "v_e": v_e, "v_a": v_a, "c_s": c_s}

    wc_1 = xr.DataArray(wc_1, coords=[k_vec], dims=["k"], attrs=attrs)
    wc_2 = xr.DataArray(wc_2, coords=[k_vec], dims=["k"], attrs=attrs)
    wc_3 = xr.DataArray(wc_3, coords=[k_vec], dims=["k"], attrs=attrs)

    return wc_1, wc_2, wc_3



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S',
                    level=logging.INFO)


def vdf_elim(vdf, e_int):
    r"""Limits the skymap distribution to the selected energy range.

    Parameters
    ----------
    vdf : xarray.Dataset
        Skymap velocity distribution to clip.
    e_int : list or float
        Energy interval boundaries (list) or energy to slice.

    Returns
    -------
    vdf_e_clipped : xarray.Dataset
        Skymap of the clipped velocity distribution.

    """

    energy = vdf.energy
    unique_etables = np.unique(vdf.energy, axis=0)
    # n_etables = 2 for older dta and 1 for newer data
    n_etables = unique_etables.shape[0]

    e_int = list(np.atleast_1d(e_int))

    # energy interval
    if len(e_int) == 2:
        e_levels = []

        for i_etable in range(n_etables):
            # loop over 1 or 2 and saves all the unique indices, i.e. max range
            lower_ = e_int[0] < unique_etables[i_etable, :]
            upper_ = unique_etables[i_etable, :] < e_int[1]
            tmp_elevels = np.where(np.logical_and(lower_, upper_))[0]
            e_levels = np.unique(np.hstack([e_levels, tmp_elevels]))

        e_levels = list(e_levels.astype(int))
        e_min = np.min(energy.data[:, e_levels])
        e_max = np.max(energy.data[:, e_levels])
        print(f"Effective eint = [{e_min:5.2f}, {e_max:5.2f}]")

    else:
        # pick closest energy level
        e_diff0 = np.abs(energy[0, :] - e_int)
        e_diff1 = np.abs(energy[1, :] - e_int)
        if np.min(e_diff0) < np.min(e_diff1):
            e_diff = e_diff0
        else:
            e_diff = e_diff1

        e_levels = int(np.where(e_diff == np.min(e_diff))[0][0])
        print(f"Effective energies alternate in time between "
              f"{energy.data[0, e_levels]:5.2f} and "
              f"{energy.data[1, e_levels]:5.2f}")

    vdf_e_clipped = ts_skymap(vdf.time.data, vdf.data[:, e_levels, ...],
                              energy=energy[:, e_levels], phi=vdf.phi,
                              theta=vdf.theta)

    energy_0 = vdf.attrs.get("energy0", unique_etables[0, :])[e_levels]
    energy_1 = vdf.attrs.get("energy1", unique_etables[0, :])[e_levels]
    esteptable = vdf.attrs.get("esteptable", np.zeros(len(vdf.time)))
    vdf_e_clipped.attrs["energy0"] = energy_0
    vdf_e_clipped.attrs["energy1"] = energy_1
    vdf_e_clipped.attrs["esteptable"] = esteptable

    return vdf_e_clipped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample
from .dot import dot
from .normalize import normalize
from .cross import cross

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eb_nrf(e_xyz, b_xyz, v_xyz, flag=0):
    """Find E and B in MP system given B and MP normal vector.

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.
    b_xyz : xarray.DataArray
        Time series of the magnetic field.
    v_xyz : xarray.DataArray
        Normal vector.
    flag : int or ndarray
        to fill.

    Returns
    -------
    out : xarray.DataArray
        to fill.

    """

    assert isinstance(flag, (int, np.ndarray)), "Invalid flag type"

    if isinstance(flag, int):
        flag_cases = ["a", "b"]
        flag_case = flag_cases[flag]
        l_direction = None

    else:
        assert np.size(flag) == 3
        l_direction = flag
        flag_case = "c"

    if flag_case == "a":
        b_data = resample(b_xyz, e_xyz).data

        n_l = b_data / np.linalg.norm(b_data, axis=0)[:, None]
        n_n = np.cross(np.cross(b_data, v_xyz), b_data)
        n_n = n_n / np.linalg.norm(n_n)[:, None]
        n_m = np.cross(n_n, n_l)  # in (vn x b) direction

    elif flag_case == "b":
        n_n = v_xyz / np.linalg.norm(v_xyz)
        n_m = normalize(cross(n_n, b_xyz.mean(dim="time")))
        n_l = cross(n_m, n_n)

    else:
        n_n = normalize(v_xyz)
        n_m = normalize(np.cross(n_n, l_direction))
        n_l = cross(n_m, n_n)

    # estimate e in new coordinates
    e_lmn = np.hstack([dot(e_xyz, vec) for vec in [n_l, n_m, n_n]])

    out = xr.DataArray(e_xyz.time.data, e_lmn, e_xyz.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect
import datetime

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def time_clip(inp, tint):
    r"""Time clip the input (if time interval is TSeries clip between start
    and stop).

    Parameters
    ----------
    inp : xarray.DataArray or xarray.Dataset
        Time series of the quantity to clip.
    tint : xarray.DataArray or ndarray or list
        Time interval can be a time series, a array of datetime64 or a list.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time clipped input.

    """

    if isinstance(inp, xr.Dataset):
        coords_data = [inp[k] for k in filter(lambda x: x != "time", inp.dims)]
        coords_data = [time_clip(inp.time, tint), *coords_data]
        out_dict = {dim: coords_data[i] for i, dim in enumerate(inp.coords)}

        for k in inp:
            if "time" in list(inp[k].coords):
                data = time_clip(inp[k], tint)
                out_dict[k] = (list(inp[k].dims), data)
            else:
                out_dict[k] = (list(inp[k].dims), inp[k])

        out = xr.Dataset(out_dict)
        out.attrs = inp.attrs
        return out

    if isinstance(tint, xr.DataArray):
        t_start, t_stop = tint.time.data[[0, -1]]

    elif isinstance(tint, np.ndarray):
        if isinstance(tint[0], datetime.datetime) \
                and isinstance(tint[-1], datetime.datetime):
            t_start, t_stop = [tint.time[0], tint.time[-1]]

        else:
            raise TypeError('Values must be in Datetime64')

    elif isinstance(tint, list):
        t_start, t_stop = iso86012datetime64(np.array(tint))

    else:
        raise TypeError("invalid tint")

    idx_min = bisect.bisect_left(inp.time.data, np.datetime64(t_start))
    idx_max = bisect.bisect_right(inp.time.data, np.datetime64(t_stop))

    coord = [inp.time.data[idx_min:idx_max]]

    if len(inp.coords) > 1:
        for k in inp.dims[1:]:
            coord.append(inp.coords[k].data)

    out = xr.DataArray(inp.data[idx_min:idx_max, ...], coords=coord,
                       dims=inp.dims, attrs=inp.attrs)

    out.time.attrs = inp.time.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np

from scipy import constants

# Local imports
from ..pyrf import estimate
from .photo_current import photo_current
from .thermal_current import thermal_current

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


class Plasma(object):
    r"""Describes plasma model consisting of several plasma components where
    each component is characterized by charge size and sign, density, mass of
    particles, temperature and drift velocity."""

    def __init__(self, name: str = "", n: Union[float, list] = None,
                 mp: Union[float, list] = None, qe: Union[int, list] = None,
                 t: Union[float, list] = None):
        r"""Setup plasma properties. They can be a single number applicable to
        all plasma components or a vector of the length equal to the number of
        plasma components.

        Parameters
        ----------
        name : str
            Name of the plasma.
        n : float or list
            Species number densities.
        mp : float or list
            Species masses in terms of proton mass.
        qe : float or list
            Species charges in terms of elementary charge.
        t : float or float
            Species temperatures.

        """

        self.name = name
        self.n = np.atleast_1d(n)
        self.mp = np.atleast_1d(mp)
        self.qe = np.atleast_1d(qe)
        self.t = np.atleast_1d(t)

        # Computes mass and charge in SI units
        self._m()
        self._q()

    def _m(self):
        self.m = self.mp * constants.proton_mass
        self.m[self.m == 0] = constants.electron_mass

    def _q(self):
        self.qe * constants.elementary_charge


class LangmuirProbe(object):
    r"""Defines either spherical, cylindrical, conical or spherical +
    cylindrical/conical probes. Probe belonging to LangmuirProbe is defined
    with properties."""

    def __init__(self, name: str, surface: str = "cluster",
                 r_sphere: Union[list, float, int] = None,
                 r_wire: Union[list, float, int] = None,
                 l_wire: Union[list, float, int] = None,
                 s_photoemission: float = None):
        r"""Setup probe properties.

        Parameters
        ----------
        name : str
            Name of the probe.
        surface : str, Optional
            Surface materials or surface photoemission in [A/m^2].
            Default is "cluster".
        r_sphere : float, Optional
            Radius of the sphere in m.
        r_wire : float or list, Optional
            Radius of the wire in m. Can be a float for wire or a list of two
            floats for a stazer.
        l_wire : float, Optional
            Length of the wire/stazer in m.
        s_photoemission : float, Optional
            Surface photoemission in A/m^2. If not given obtain from surface
            type.

        Raises
        ------
        AssertionError : if r_wire as more than 2 elements.

        """

        message = "The wire radius should be a numeric vector of " \
                  "length 1 (wire) or 2 (stazer)."
        assert not r_wire or np.atleast_1d(r_wire) <= 2, message

        self.name = name
        self.surface = surface
        self.r_sphere = r_sphere
        self.r_wire = np.atleast_1d(r_wire)
        self.l_wire = l_wire
        self.s_photoemission = s_photoemission

        # Probe type according to the specified parameters: Sphere + Wire or
        # Wire or Sphere
        self._get_probe_type()

        # Get probe area
        self._get_probe_area()

        # Get probe capacitance
        self._get_probe_capa()

    def _get_probe_type(self):
        r"""Define probe type according to the specified parameters."""
        if self.r_sphere and self.r_wire and self.l_wire:
            self.probe_type = "sphere+wire"
        elif self.r_wire and self.l_wire:
            self.probe_type = "wire"
        elif self.r_sphere:
            self.probe_type = "sphere"
        else:
            self.probe_type = None

    def _get_probe_area(self):
        r"""Computes probe area."""
        a_sphere_sunlit, a_sphere_total = [0, 0]
        a_wire_sunlit, a_wire_total = [0, 0]

        if isinstance(self.r_sphere, (float, int)):
            a_sphere_sunlit = np.pi * self.r_sphere ** 2
            a_sphere_total = 4 * a_sphere_sunlit

        if self.r_wire and isinstance(self.r_wire, (float, int, list)) \
                and self.l_wire and isinstance(self.l_wire, (float, int)):
            a_wire_sunlit = 2 * np.mean(self.r_wire) * self.l_wire
            a_wire_total = np.pi * a_wire_sunlit

        self.area = {"sphere": a_sphere_total, "wire": a_wire_total,
                     "total": a_sphere_total + a_wire_total,
                     "sunlit": a_sphere_sunlit + a_wire_sunlit}
        self.area["total_sunlit"] = self.area["total"] / self.area["sunlit"]
        self.area["sunlit_total"] = 1 / self.area["total_sunlit"]

    def _get_probe_capa(self):
        r"""Computes probe capacitance.

        Raises
        ------
        ValueError : if length > radius.

        See Also
        --------
        pyrfu.pyrf.estimate.py

        """

        c_wire = 0
        c_sphere = estimate("capacitance_sphere", self.r_sphere)

        if self.r_wire and isinstance(self.r_wire, (float, int, list)) \
                and self.l_wire and isinstance(self.l_wire, (float, int)):
            if self.l_wire > 10 * list([self.r_wire]):
                c_wire = estimate("capacitance_wire", np.mean(self.r_wire),
                                  self.l_wire)
            elif self.l_wire > list(self.r_wire):
                c_wire = estimate("capacitance_cylinder", np.mean(self.r_wire),
                                  self.l_wire)
            else:
                raise ValueError("estimate of capacitance for cylinder "
                                 "requires length > radius")

        self.capacitance = np.sum([c_sphere, c_wire])

    def _get_probe_surface_photoemission(self):
        r"""Computes (or get) surface photo emission."""
        if self.surface:
            self.s_photoemission = self.s_photoemission
        else:
            self.s_photoemission = photo_current(1., 0., 1., self.surface)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pdb

# 3rd party imports
import numpy as np
import xarray as xr

from cdflib import cdfread

# Local imports
from ..pyrf import datetime642iso8601

from .list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _check_time(proton_phxtof, proton_extof):
    data_size = [len(proton_phxtof), len(proton_extof)]

    if data_size[0] == data_size[1]:
        # identify mismatching timesteps
        cond = proton_phxtof.time.data != proton_extof.time.data
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            e_inds = []
            ph_inds = []
            for t in range(data_size[1]):
                dt_dummy = np.min(
                    np.abs(proton_extof.time.data[t] - proton_extof.time.data))
                t_ind = np.argmin(np.abs(proton_extof.time.data[t]
                                         - proton_extof.time.data))
                if dt_dummy == 0:
                    e_inds.append(t)
                    if not ph_inds:
                        ph_inds.append(t)
                    else:
                        ph_inds.append(t_ind)

            time_data = proton_extof.time.data[e_inds]
            phxtof_data = proton_phxtof.data[ph_inds, ...]
            extof_data = proton_extof.data[e_inds, ...]
        else:
            time_data = proton_extof.time.data
            phxtof_data = proton_phxtof.data
            extof_data = proton_extof.data

    elif data_size[0] > data_size[1]:
        cond = proton_phxtof.time.data[:data_size[1]] != proton_extof.time.data
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            e_inds = []
            ph_inds = []
            for t in range(data_size[1]):
                dt_dummy = np.min(np.abs(proton_extof.time.data[t]
                                         - proton_phxtof.time.data))
                t_ind = np.argmin(np.abs(proton_extof.time.data[t]
                                         - proton_phxtof.time.data))
                if dt_dummy == 0:
                    e_inds.append(t)
                    if not ph_inds:
                        ph_inds.append(t)
                    else:
                        ph_inds.append(t_ind)

            time_data = proton_extof.time.data[e_inds]
            phxtof_data = proton_phxtof.data[ph_inds, ...]
            extof_data = proton_extof.data[e_inds, ...]
        else:
            time_data = proton_extof.time.data
            phxtof_data = proton_phxtof.data[:data_size[1], ...]
            extof_data = proton_extof.data

    elif data_size[0] < data_size[1]:
        cond = proton_phxtof.time.data != proton_extof.time.data[:data_size[0]]
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            e_inds = []
            ph_inds = []
            for t in range(data_size[0]):
                dt_dummy = np.min(np.abs(proton_phxtof.time.data[t]
                                         - proton_phxtof.time.data))
                t_ind = np.argmin(np.abs(
                    proton_phxtof.time.data[t] - proton_phxtof.time.data))
                if dt_dummy == 0:
                    ph_inds.append(t)
                    if not ph_inds:
                        e_inds.append(t)
                    else:
                        e_inds.append(t_ind)

            time_data = proton_phxtof.time.data[ph_inds]
            phxtof_data = proton_phxtof.data[ph_inds, ...]
            extof_data = proton_extof.data[e_inds, ...]

        else:
            time_data = proton_phxtof.time.data
            phxtof_data = proton_phxtof.data
            extof_data = proton_extof.data[:data_size[0], ...]

    else:
        raise ValueError

    return time_data, phxtof_data, extof_data


def _get_energy_dplus_dminus(eis_allt, data_path):
    tint = list(datetime642iso8601(eis_allt.time.data[[0, -1]]))

    name_ = eis_allt.t0.attrs["FIELDNAM"]

    mms_id = int(name_.split("_")[0][-1])

    var = {"inst": "epd-eis", "lev": "l2"}

    if "brst" in name_:
        var["tmmode"] = "brst"
    else:
        var["tmmode"] = "srvy"

    var["dtype"] = name_.split("_")[-5]

    files = list_files(tint, mms_id, var, data_path=data_path)

    with cdfread.CDF(files[0]) as file:
        d_plus = file.varget(eis_allt.t0.energy.attrs["DELTA_PLUS_VAR"])
        d_minus = file.varget(eis_allt.t0.energy.attrs["DELTA_MINUS_VAR"])

    return d_plus, d_minus


def eis_combine_proton_spec(phxtof_allt, extof_allt):
    r"""Combine ExTOF and PHxTOF proton energy spectra into a single combined
    Dataset.

    Parameters
    ----------
    phxtof_allt : xarray.Dataset
        Dataset containing the PHxTOF energy spectrum of the 6 telescopes.
    extof_allt : xarray.Dataset
        Dataset containing the ExTOF energy spectrum of the 6 telescopes.

    Returns
    -------
    comb_allt : xarray.Dataset
        Dataset containing the combined PHxTOF and ExTOF energy spectrum of
        the 6 telescopes.

    """

    scopes_phxtof = list(filter(lambda x: x[0] == "t", phxtof_allt))
    scopes_extof = list(filter(lambda x: x[0] == "t", extof_allt))
    assert scopes_extof == scopes_phxtof

    data_path = extof_allt.attrs["data_path"]
    dp_phxtof, dm_phxtof = _get_energy_dplus_dminus(phxtof_allt, data_path)
    dp_extof, dm_extof = _get_energy_dplus_dminus(extof_allt, data_path)

    out_keys = list(filter(lambda x: x not in scopes_extof, extof_allt))
    out_dict = {k: extof_allt[k] for k in out_keys if k != "sector"}

    comb_en_low, comb_en_hig = [None] * 2

    time_sect, phxtof_sect, extof_sect = _check_time(phxtof_allt["sector"],
                                                     extof_allt["sector"])
    sect = xr.DataArray(extof_sect, coords=[time_sect], dims=["time"])

    time_spin, phxtof_spin, extof_spin = _check_time(phxtof_allt["spin"],
                                                     extof_allt["spin"])

    spin = xr.DataArray(extof_spin, coords=[time_sect], dims=["time"])

    for scope in scopes_phxtof:
        proton_phxtof = phxtof_allt[scope]
        proton_extof = extof_allt[scope]

        time_data, phxtof_data, extof_data = _check_time(proton_phxtof,
                                                         proton_extof)

        en_phxtof, en_extof = [proton_phxtof.energy.data,
                               proton_extof.energy.data]
        idx_phxtof = np.where(en_phxtof < en_extof[0])[0]
        cond_ = np.logical_and(en_phxtof > en_extof[0],
                               en_phxtof < en_phxtof[-1])
        idx_phxtof_cross = np.where(cond_)[0]

        idx_extof_cross = np.where(en_extof < en_phxtof[-2])[0]
        idx_extof = np.where(en_extof > en_phxtof[-2])[0]

        n_phxtof = idx_phxtof.size
        n_phxtof_cross = idx_phxtof_cross.size
        n_extof = idx_extof.size

        n_en = n_phxtof + n_phxtof_cross + n_extof

        comb_en, comb_en_low, comb_en_hig = [np.zeros(n_en) for _ in range(3)]

        comb_array = np.zeros((len(time_data), n_en))
        comb_array[:, :n_phxtof] = phxtof_data[:, idx_phxtof]
        comb_en[:n_phxtof] = en_phxtof[idx_phxtof]
        comb_en_low[:n_phxtof] = comb_en[:n_phxtof] - dm_phxtof[idx_phxtof]
        comb_en_hig[:n_phxtof] = comb_en[:n_phxtof] + dp_phxtof[idx_phxtof]

        for (i, i_phx), i_ex in zip(enumerate(idx_phxtof_cross),
                                    idx_extof_cross):
            idx_ = n_phxtof + i
            comb_array[:, idx_] = np.nanmean(np.vstack([phxtof_data[:, i_phx],
                                                        extof_data[:, i_ex]]),
                                             axis=0)
            comb_en_low[idx_] = np.nanmin([en_phxtof[idx_] - dm_phxtof[idx_],
                                           en_extof[i] - dm_extof[i]])
            comb_en_hig[idx_] = np.nanmax([en_phxtof[idx_] + dp_phxtof[idx_],
                                           en_extof[i] + dp_extof[i]])
            comb_en[idx_] = np.sqrt(comb_en_low[idx_] * comb_en_hig[idx_])

        comb_array[:, -n_extof:] = extof_data[:, idx_extof]
        comb_en[-n_extof:] = en_extof[idx_extof]
        comb_en_low[-n_extof:] = en_extof[idx_extof] - dm_extof[idx_extof]
        comb_en_hig[-n_extof:] = en_extof[idx_extof] + dp_extof[idx_extof]

        out_dict[scope] = xr.DataArray(comb_array,
                                       coords=[time_data, comb_en],
                                       dims=["time", "energy"])

    out_dict["sector"] = sect
    out_dict["spin"] = spin
    out_dict["energy_dminus"] = comb_en_low
    out_dict["energy_dplus"] = comb_en_hig

    comb_allt = xr.Dataset(out_dict)

    return comb_allt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.colors as mcolors

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"

plt.style.use("seaborn-ticks")


def plot_spectr(axis, inp, yscale: str = "linear", cscale: str = "linear",
                clim: list = None, cmap: str = "", colorbar: bool = True,
                **kwargs):
    r"""Plot a spectrogram using pcolormesh.

    Parameters
    ----------
    axis : matplotlib.pyplot.subplotsaxes
        Target axis to plot. If None create a new figure.
    inp : xarray.DataArray
        Input 2D data to plot.
    yscale : {"linear", "log"}, Optional
        Y-axis scaling. Default is "" (linear).
    cscale : {"linear", "log"}, Optional
        C-axis scaling. Default is "" (linear).
    clim : list, Optional
        C-axis bounds. Default is None (autolim).
    cmap : str, Optional
        Colormap. Default is "jet".
    colorbar : bool, Optional
        Flag for colorbar. Set to False to hide.

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments.

    Returns
    -------
    fig : figure
        Figure with axis.
    axs : matplotlib.pyplot.subplotsaxes
        Axis with spectrum.
    caxs : matplotlib.pyplot.subplotsaxes
        Only if colorbar is True.

    """

    if axis is None:
        fig, axis = plt.subplots(1)
    else:
        fig = plt.gcf()

    if not cmap:
        cmap = "jet"

    if cscale == "log":
        if clim is not None and isinstance(clim, list):
            options = dict(norm=mcolors.LogNorm(vmin=clim[0], vmax=clim[1]),
                           cmap=cmap)
        else:
            options = dict(norm=mcolors.LogNorm(), cmap=cmap)
    else:
        if clim is not None and isinstance(clim, list):
            options = dict(cmap=cmap, vmin=clim[0], vmax=clim[1])
        else:
            options = dict(cmap=cmap, vmin=None, vmax=None)

    x_data, y_data = [inp.coords[inp.dims[0]], inp.coords[inp.dims[1]]]

    image = axis.pcolormesh(x_data, y_data, inp.data.T, rasterized=True,
                            shading="auto", **options)

    if x_data.dtype == '<M8[ns]':
        locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
        formatter = mdates.ConciseDateFormatter(locator)
        axis.xaxis.set_major_locator(locator)
        axis.xaxis.set_major_formatter(formatter)

    if yscale == "log":
        axis.set_yscale("log")

    if colorbar:
        if kwargs.get("pad"):
            pad = kwargs["pad"]
        else:
            pad = 0.01

        pos = axis.get_position()
        cax = fig.add_axes([pos.x0+pos.width+pad, pos.y0, 0.01, pos.height])
        fig.colorbar(mappable=image, cax=cax, ax=axis)

        out = (axis, cax)
    else:
        out = axis

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pdb

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"

def _idx_closest(lst0, lst1):
    return [(np.abs(np.asarray(lst0) - k)).argmin() for k in lst1]


def eis_skymap_combine_sc(skymaps):
    r"""Generate composite skymap from the EIS sensors across the MMS
    spacecraft.

    Parameters
    ----------
    skymaps : list of xarray.DataArray
        Skymap distribution for all spacecraft.

    Returns
    -------
    out : xarray.Dataset
        Composite skymap distribution

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_pad,
    pyrfu.mms.eis_spec_combine_sc, pyrfu.mms.eis_spec_combine_sc

    """

    # Determine spacecraft with smallest number of time steps to use as
    # reference spacecraft
    time_size = [len(probe.time.data) for probe in skymaps]
    ref_sc_time_size, ref_sc_loc = [np.min(time_size), np.argmin(time_size)]
    ref_probe = skymaps[ref_sc_loc]

    # Define common energy grid across EIS instruments
    n_en_chans = [probe.energy.shape[1] for probe in skymaps]
    size_en, loc_ref_en = [np.min(n_en_chans), np.argmin(n_en_chans)]
    ref_energy = skymaps[loc_ref_en].energy.data[0, :]

    pdb.set_trace()

    energy_data, e_plus, e_minu = [[], [], []]
    for probe in skymaps:
        idx = _idx_closest(probe.energy.data[0, :], ref_energy)
        energy_data.append(probe.energy.data[0, idx])
        e_minu.append(probe.attrs["energy_dminus"][idx])
        e_plus.append(probe.attrs["energy_dplus"][idx])

    energy_data = np.stack(energy_data)
    common_energy = np.nanmean(energy_data, axis=0)
    common_energy = np.tile(common_energy, (ref_sc_time_size, 1))

    #
    e_minu = np.stack(e_minu)
    e_plus = np.stack(e_plus)
    common_minu = np.nanmean(e_minu, axis=0)
    common_plus = np.nanmean(e_plus, axis=0)

    # Use azimuthal and elevation angle from reference spacecraft (in
    # practice they are the same for all spacecraft)
    phi = ref_probe.phi.data
    theta = ref_probe.theta.data

    allmms_skymap = np.zeros([ref_sc_time_size, size_en, phi.shape[1],
                              len(theta), len(skymaps)])

    for p, skymap in enumerate(skymaps):
        idx_en = _idx_closest(skymap.energy.data[0, :], common_energy[0, :])
        allmms_skymap[..., p] = skymap.data[:ref_sc_time_size, idx_en, ...]

    # Average the four spacecraft
    allmms_skymap_avg = np.nanmean(allmms_skymap, axis=-1)

    # Create combined skymap
    out_dict = {"time": ref_probe.time.data,
                "idx0": range(common_energy.shape[1]),
                "idx1": range(phi.shape[1]), "idx2": range(len(theta)),
                "data": (["time", "idx0", "idx1", "idx2"], allmms_skymap_avg),
                "energy": (["time", "idx0"], common_energy),
                "phi": (["time", "idx1"], phi), "theta": (["idx2"], theta)}

    out = xr.Dataset(out_dict)

    out.attrs["energy_dminus"] = common_minu
    out.attrs["energy_dplus"] = common_plus

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import numpy as np


from pyrfu import mms, pyrf, models, dispersion

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os
import subprocess

# Local imports
from .mms_config import CONFIG
from .list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def copy_files(var, tint, mms_id, target_dir: str = "./data/"):
    r"""Copy files from NAS24 to the target path

    Parameters
    ----------
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument.
            * var["tmmode"] : data rate.
            * var["lev"] : data level.
            * var["dtype"] : data type.
    tint : list of str
        Time interval.
    mms_id : str or int
        Index of the spacecraft.
    target_dir : str, Optional
        Target path. Default is './data/'.

    """

    mms_path = CONFIG["local_data_dir"] + "/"

    files = list_files(tint, mms_id, var)

    for file in files:
        relative_path = os.path.split(file)[0].replace(mms_path, "")
        path = os.path.join(target_dir, relative_path)
        target_file = os.path.join(path, os.path.split(file)[1])

        if not os.path.exists(path):
            os.makedirs(path)

        p = subprocess.Popen(f"cp {file} {target_file}",
                             stdout=subprocess.PIPE,
                             shell=True)
        (_, _) = p.communicate()

        # This makes the wait possible
        _ = p.wait()

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def histogram(inp, bins: int = 100, normed: bool = True):
    r"""Computes 1D histogram of the inp with bins bins

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input scalar variable.

    bins : int, Optional
        Number of bins. Default is 100.

    normed : bool, Optional
        Normalize the PDF. Default is True.

    Returns
    -------
    out : xarray.DataArray
        1D distribution of the input time series.

    """

    hist, bins = np.histogram(inp.data, bins=bins, normed=normed)
    bin_center = (bins[1:] + bins[:-1]) * 0.5

    out = xr.DataArray(hist, coords=[bin_center], dims=["bins"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .iso86012datetime64 import iso86012datetime64
from .datetime642iso8601 import datetime642iso8601

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def extend_tint(tint, ext: list = None):
    r"""Extends time interval.

    Parameters
    ----------
    tint : list of str
        Reference time interval to extend.
    ext : list of float or list of int
        Number of seconds to extend time interval
        [left extend, right extend].

    Returns
    -------
    tint_new : list of str
        Extended time interval.

    Examples
    --------
    >>> from pyrfu import pyrf

    Time interval

    >>> tints = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]

    Load spacecraft position

    >>> tints_long = pyrf.extend_tint(tint, [-100, 100])

    """

    if ext is None:
        ext = [-60, 60]

    # Convert extension to timedelta64 in s units
    ext = np.array(ext).astype("timedelta64[s]")

    # Original time interval to datetime64 format in ns units
    tint_ori = iso86012datetime64(np.array(tint))

    # New time interval in iso 8601 format
    tint_new = list(datetime642iso8601(tint_ori + ext))

    return tint_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np

from scipy.special import erf
from scipy.constants import elementary_charge, Boltzmann

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _spherical_body(z, u, x_, i_p):
    j_thermal = np.zeros(u.shape)

    if z > 0:
        j_thermal[u >= 0] = i_p * np.exp(-x_[u >= 0])
        j_thermal[u < 0] = i_p * (1 - x_[u < 0])
    elif z < 0:
        j_thermal[u >= 0] = i_p * (1 + x_[u >= 0])
        j_thermal[u < 0] = i_p * np.exp(x_[u < 0])

    return j_thermal


def _cylindrical_body(z, u, x_, i_p):
    sq = np.zeros(u.shape)
    j_thermal = np.zeros(u.shape)

    sq[u < 0] = np.sqrt(abs(-x_[u < 0]))
    sq[u >= 0] = np.sqrt(abs(+x_[u >= 0]))
    erfv = erf(sq)

    if z > 0:
        j_thermal[u >= 0] = i_p * np.exp(-x_[u >= 0])
        c_0 = (2.0 / np.sqrt(np.pi)) * sq[u < 0]
        c_1 = np.exp(-x_[u < 0]) * (1.0 - erfv[u < 0])
        j_thermal[u < 0] = i_p * (c_0 + c_1)
    elif z < 0:
        j_thermal[u < 0] = i_p * np.exp(x_[u < 0])

        c_0 = (2.0 / np.sqrt(np.pi)) * sq[u >= 0]
        c_1 = np.exp(x_[u >= 0]) * (1.0 - erfv[u >= 0])
        j_thermal[u >= 0] = i_p * (c_0 + c_1)

    return j_thermal


def thermal_current(n: float, t: float, m: float, v: float, z: float,
                    u: Union[float, np.ndarray], a: float,
                    p_type: str) -> Union[float, np.ndarray]:
    r"""Calculates the thermal probe current to/from a cylindrical or
    spherical body, e.g. a Langmuir probe or the a spherical (cylindrical) S/C.

    Parameters
    ----------
    n : float
        Number density [m^3].
    t : float
        Temperature [K]
    m : float
        Mass [kg].
    z : {-1, 1}
        Charge
    v : float
        Velocity of the body with respect to the plasma [m/s].
    u : float or numpy.ndarray
        Body potential [V]
    a : float
        Area of body [m^2].
    p_type : {"sphere", "cylinder"}
        Probe type.

    Returns
    -------
    j_thermal : float or numpy.ndarray

    """

    assert p_type.lower() in ["sphere", "cylinder"]

    u = np.atleast_1d(u)

    # If zero density return zero current
    if n == 0 or t == 0:
        return

    # Is the body moving with a velocity, V, with respect to the plasma ?
    # Criteria set such that it is considered important if V > 0.1 * V_th.
    if v < 0.1 * np.sqrt(Boltzmann * t / m):
        # Ratio of potential to thermal energy.
        x_ = elementary_charge * u / (Boltzmann * t)

        # Total current to/from body.
        i_p = np.sqrt(t * Boltzmann / (2.0 * np.pi * m))
        i_p *= a * n * elementary_charge

    else:
        x_ = (elementary_charge / (m * v ** 2 / 2 + Boltzmann * t)) * u
        i_p = np.sqrt(v ** 2 / 16 + t * Boltzmann / (2.0 * np.pi * m))
        i_p *= a * n * elementary_charge

    if p_type == "sphere":
        # Spherical body case.
        j_thermal = _spherical_body(z, u, x_, i_p)

    else:
        # Cylindrical body case.
        j_thermal = _cylindrical_body(z, u, x_, i_p)

    return j_thermal



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from cdflib import CDF, cdfepoch

# Local imports
from ..pyrf import datetime642iso8601, iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _get_epochs(file, cdf_name, tint):
    depend0_key = file.varattsget(cdf_name)["DEPEND_0"]

    out = {"data": file.varget(depend0_key,
                               starttime=tint[0], endtime=tint[1])}

    if file.varinq(depend0_key)["Data_Type_Description"] == "CDF_TIME_TT2000":
        out["data"] = cdfepoch.to_datetime(out["data"], to_np=True)

    out["atts"] = file.varattsget(depend0_key)

    return out


def _get_depend_attributes(file, depend_key):
    attributes = file.varattsget(depend_key)

    # Remove spaces in label
    try:
        attributes["LABLAXIS"] = attributes["LABLAXIS"].replace(" ", "_")

        if attributes["LABLAXIS"] == "Diffential_energy_channels":
            attributes["LABLAXIS"] = "Differential_energy_channels"

    except (KeyError, AttributeError):
        attributes["LABLAXIS"] = "comp"

    return attributes


def _get_depend(file, cdf_name, tint, dep_num=1):
    out = {}

    try:
        depend_key = file.varattsget(cdf_name)[f"DEPEND_{dep_num:d}"]
    except KeyError:
        depend_key = file.varattsget(cdf_name)[f"REPRESENTATION_{dep_num:d}"]

    if depend_key == "x,y,z":
        out["data"] = np.array(depend_key.split(","))

        out["atts"] = {"LABLAXIS": "comp"}
    else:
        try:
            out["data"] = file.varget(depend_key, starttime=tint[0],
                                      endtime=tint[1])
        except IndexError:
            out["data"] = file.varget(depend_key)

        out["data"] = file.varget(depend_key)

        if len(out["data"]) == 1:
            out["data"] = out["data"][0]

        if len(out["data"]) == 4 and all(out["data"] == ["x", "y", "z", "r"]):
            out["data"] = out["data"][:-1]

        elif out["data"].ndim == 2:
            if len(out["data"].flatten()) == 3:
                out["data"] = out["data"].flatten()
            else:
                try:
                    out["data"] = out["data"][0, :]
                except IndexError:
                    pass

        out["atts"] = _get_depend_attributes(file, depend_key)

    return out


def get_ts(file_path, cdf_name, tint):
    r"""Reads field named cdf_name in file and convert to time series.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.
    tint : list of str
        Time interval.

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable in the selected time interval.

    """

    # Convert time interval to epochs
    tint = list(datetime642iso8601(iso86012datetime64(np.array(tint))))
    tint = list(map(cdfepoch.parse, tint))

    out_dict = {}
    time, depend_1, depend_2, depend_3 = [{}, {}, {}, {}]

    with CDF(file_path) as file:
        attrs_ = file.varattsget(cdf_name)
        out_dict["atts"] = attrs_

        assert "DEPEND_0" in attrs_ and "epoch" in attrs_["DEPEND_0"].lower()

        time = _get_epochs(file, cdf_name, tint)

        if "DEPEND_1" in attrs_ or "REPRESENTATION_1" in attrs_:
            depend_1 = _get_depend(file, cdf_name, tint, 1)

        elif "afg" in cdf_name or "dfg" in cdf_name:
            depend_1 = {"data": ["x", "y", "z"], "atts": {"LABLAXIS": "comp"}}

        if "DEPEND_2" in attrs_ or "REPRESENTATION_2" in attrs_:
            depend_2 = _get_depend(file, cdf_name, tint, 2)

            if depend_2["atts"]["LABLAXIS"] == depend_1["atts"]["LABLAXIS"]:
                depend_1["atts"]["LABLAXIS"] = "rcomp"
                depend_2["atts"]["LABLAXIS"] = "ccomp"

        if "DEPEND_3" in attrs_ or "REPRESENTATION_3" in attrs_:
            if "REPRESENTATION_3" in attrs_:
                assert out_dict["atts"]["REPRESENTATION_3"] != "x,y,z"

            depend_3 = _get_depend(file, cdf_name, tint, 3)

            if depend_3["atts"]["LABLAXIS"] == depend_2["atts"]["LABLAXIS"]:
                depend_2["atts"]["LABLAXIS"] = "rcomp"
                depend_3["atts"]["LABLAXIS"] = "ccomp"

        if "sector_mask" in cdf_name:
            cdf_name_mask = cdf_name.replace("sector_mask", "intensity")
            depend_1_key = file.varattsget(cdf_name_mask)["DEPEND_1"]

            depend_1["data"] = file.varget(depend_1_key)
            depend_1["atts"] = file.varattsget(depend_1_key)

            depend_1["atts"]["LABLAXIS"] = depend_1["atts"][
                "LABLAXIS"].replace(" ", "_")

        if "edp_dce_sensor" in cdf_name:
            depend_1["data"] = ["x", "y", "z"]
            depend_1["atts"] = {"LABLAXIS": "comp"}

        out_dict["data"] = file.varget(cdf_name, starttime=tint[0],
                                       endtime=tint[1])

        if out_dict["data"].ndim == 2 and out_dict["data"].shape[1] == 4:
            out_dict["data"] = out_dict["data"][:, :-1]

    if out_dict["data"].ndim == 2 and not depend_1:
        depend_1["data"] = np.arange(out_dict["data"].shape[1])
        depend_1["atts"] = {"LABLAXIS": "idx"}

    if time and not depend_1 and not depend_2 and not depend_3:
        dims = ["time"]
        coords_data = [time["data"]]
        coords_atts = [time["atts"]]

    elif time and depend_1 and not depend_2 and not depend_3:
        dims = ["time", depend_1["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"]]
        coords_atts = [time["atts"], depend_1["atts"]]

    elif time and depend_1 and depend_2 and not depend_3:
        if depend_1["atts"]["LABLAXIS"] == depend_2["atts"]["LABLAXIS"]:
            depend_1["atts"]["LABLAXIS"] = "rcomp"
            depend_2["atts"]["LABLAXIS"] = "ccomp"

        dims = ["time", depend_1["atts"]["LABLAXIS"],
                depend_2["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"], depend_2["data"]]
        coords_atts = [time["atts"], depend_1["atts"], depend_2["atts"]]

    elif time and depend_1 and depend_2 and depend_3:
        if depend_2["atts"]["LABLAXIS"] == depend_3["atts"]["LABLAXIS"]:
            depend_2["atts"]["LABLAXIS"] = "rcomp"
            depend_3["atts"]["LABLAXIS"] = "ccomp"

        dims = ["time", depend_1["atts"]["LABLAXIS"],
                depend_2["atts"]["LABLAXIS"], depend_3["atts"]["LABLAXIS"]]
        coords_data = [time["data"], depend_1["data"], depend_2["data"],
                       depend_3["data"]]
        coords_atts = [time["atts"], depend_1["atts"], depend_2["atts"],
                       depend_3["atts"]]

    else:
        raise NotImplementedError

    out = xr.DataArray(out_dict["data"], coords=coords_data, dims=dims,
                       attrs=out_dict["atts"])

    for dim, coord_atts in zip(dims, coords_atts):
        out[dim].attrs = coord_atts

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def solid_angle(inp0, inp1, inp2):
    r"""Calculates the solid angle of three vectors making up a triangle
    in a unit sphere with the sign taken into account.

    Parameters
    ----------
    inp0 : ndarray
        First vector.
    inp1 : ndarray
        Second vector.
    inp2 : ndarray
        Third vector.

    Returns
    -------
    angle : float
        Solid angle.

    """

    # Calculate the smaller angles between the vectors around origin
    acos_12 = np.arccos(np.sum(inp2 * inp1))
    acos_02 = np.arccos(np.sum(inp0 * inp2))
    acos_01 = np.arccos(np.sum(inp1 * inp0))

    # Calculate the angles in the spherical triangle (Law of Cosines)
    alpha = np.arccos((np.cos(acos_12) - np.cos(acos_02) * np.cos(acos_01))
                      / (np.sin(acos_02) * np.sin(acos_01)))
    beta = np.arccos((np.cos(acos_02) - np.cos(acos_12) * np.cos(acos_01)) /
                     (np.sin(acos_12) * np.sin(acos_01)))
    gamma = np.arccos((np.cos(acos_01) - np.cos(acos_02) * np.cos(acos_12))
                      / (np.sin(acos_02) * np.sin(acos_12)))

    # Calculates the Surface area on the unit sphere (solid angle)
    angle = (alpha + beta + gamma - np.pi)
    # Calculate the sign of the area
    var = np.cross(inp2, inp1)
    div = np.sum(var * inp0)
    sgn = np.sign(div)

    # Solid angle with sign taken into account
    angle = sgn * angle

    return angle



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np
from geopack import geopack
from matplotlib.patches import Wedge

# Local imports
from ..pyrf import magnetosphere, datetime642unix, iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S',
                    level=logging.INFO)


def _add_earth(ax=None, **kwargs):
    theta1, theta2 = 90., 270.
    nightside_ = Wedge((0., 0.), 1., theta1, theta2, fc="k", ec="k", **kwargs)
    dayside_ = Wedge((0., 0.), 1., theta2, theta1, fc="w", ec="k", **kwargs)
    for wedge in [nightside_, dayside_]:
        ax.add_artist(wedge)
    return [nightside_, dayside_]


def _add_field_lines(ax, tint):
    # Get dipole axis at begin of the time interval
    ut = datetime642unix(iso86012datetime64(np.array(tint)))[0]
    ps = geopack.recalc(ut)

    x_lines_m, y_lines_m, z_lines_m = [[], [], []]
    x_lines_p, y_lines_p, z_lines_p = [[], [], []]

    xx_gsm, zz_gsm = np.meshgrid(np.linspace(-30, 6, 19),
                                 np.linspace(-5, 5, 10))
    xx_gsm = np.reshape(xx_gsm, (xx_gsm.size,))
    zz_gsm = np.reshape(zz_gsm, (zz_gsm.size,))

    for x_gsm, z_gsm in zip(xx_gsm, zz_gsm):
        y_gsm = 0
        _, _, _, xx, yy, zz = geopack.trace(x_gsm, y_gsm, z_gsm, dir=-1,
                                            rlim=100,
                                            r0=.99999, parmod=2, exname='t89',
                                            inname='igrf', maxloop=10000)
        x_lines_m.append(xx)
        z_lines_m.append(zz)

        _, _, _, xx, yy, zz = geopack.trace(x_gsm, y_gsm, z_gsm, dir=1,
                                            rlim=100,
                                            r0=.99999, parmod=2, exname='t89',
                                            inname='igrf', maxloop=10000)
        x_lines_p.append(xx)
        z_lines_p.append(zz)

    for xx, zz in zip(x_lines_m, z_lines_m):
        ax.plot(xx, zz, color="k", linewidth=.2)

    for xx, zz in zip(x_lines_p, z_lines_p):
        ax.plot(xx, zz, color="k", linewidth=.2)

    return


def plot_magnetosphere(ax, tint, colors: list = None,
                       field_lines: bool = True):
    r"""Plot magnetopause, bow shock and earth.

    Parameters
    ----------
    ax : matplotlib.pyplot.subplotsaxes
        Axis to plot.
    tint : list of str
        Time interval.
    colors : list, Optional
        Colors of the magnetopause and the bow show.
        Default use ["tab:blue", "tab:red"]

    Returns
    -------
    ax : matplotlib.pyplot.subplotsaxes
        Axis.

    """

    # Compute Magnetopause
    if colors is None:
        colors = ["tab:blue", "tab:green"]

    x_mp, y_mp = magnetosphere("mp_shue1998", tint)

    # Compute bow show
    x_bs, y_bs = magnetosphere("bs", tint)

    # Plot
    ax.plot(np.hstack([x_mp, np.flip(x_mp)]),
            np.hstack([y_mp, np.flip(-y_mp)]),
            color=colors[0], label="Magnetopause")
    ax.plot(np.hstack([x_bs, np.flip(x_bs)]),
            np.hstack([y_bs, np.flip(-y_bs)]),
            color=colors[1], label="Bow Shock")
    _add_earth(ax)

    if field_lines:
        logging.info("Computing field lines using T89 model...")
        _add_field_lines(ax, tint)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eis_energy_corr(vdf, gamma):
    r"""Compute energy correction factor as

    .. math::

        \\left ( \\frac{E_{eff}}{E_{g}} \\right ) = \frac{1}{\\gamma - 1}
        \\frac{r^{\\gamma - 1} - r^{-\\gamma + 1}}{r^{1} - r^{-1}}

    Parameters
    ----------
    vdf : xarray.Dataset
        Ion skymap distribution.
    gamma : float or array_like
        Power law index.

    Returns
    -------
    corr : np.ndarray
        Energy correction factor

    References
    ----------
    .. [1]  Elena A. Kronberg, Patrick W. Daly, and Esa Vilenius (2021),
            Calibration Report of the RAPID Measurements in the Cluster
            Science Archive (CSA), Version 7.0,
            url: https://caa.esac.esa.int/documents/CR/CAA_EST_CR_RAP_v70.pdf

    """

    # Unpack lower and upper bounds of the energy channels
    e_lower = vdf.energy_dminus.data
    e_upper = vdf.energy_dplus.data

    if isinstance(gamma, list):
        assert len(gamma) == len(e_lower)

    # Compute upper to lower bound energy ratio
    r = e_upper / e_lower

    # Compute correction factor
    pow_ = gamma - 1
    corr = pow_ ** -1 * (r ** pow_ - r ** -pow_) / (r ** 1 - r ** -1)
    corr = corr ** (1 / pow_)

    # Correct energies
    out = vdf.copy()
    out.energy.data *= corr

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .plot_line import plot_line
from .plot_spectr import plot_spectr
from .mms_pl_config import mms_pl_config
from .pl_scatter_matrix import pl_scatter_matrix
from .zoom import zoom
from .colorbar import colorbar
from .make_labels import make_labels
from .pl_tx import pl_tx
from .add_position import add_position
from .plot_magnetosphere import plot_magnetosphere
from .plot_projection import plot_projection
from .plot_surf import plot_surf
from .span_tint import span_tint
from .plot_clines import plot_clines
from .plot_sitl_overview import plot_sitl_overview
from .plot_ang_ang import plot_ang_ang
from .plot_heatmap import plot_heatmap
from .annotate_heatmap import annotate_heatmap


__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings
import itertools

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eis_correlation(flux_s0, flux_s1):
    r"""Computes EIS correlaion matrix between specie 1 and specie 2,
    as described in [1]_ and [2]_ .

    Parameters
    ----------
    flux_s0 : xarray.DataArray
        Time series of the omni-directional differential particle flux of the
        first specie.
    flux_s1 : xarray.DataArray
        Time series of the omni-directional differential particle flux of the
        second specie.

    Returns
    -------
    energy_s0 : numpy.ndarray
        Energy levels of the first species.
    energy_s1 : numpy.ndarray
        Energy levels of the second species.
    corr_mat : numpy.ndarray
        Correlation matrix.

    References
    ----------
    .. [1]  Mitchell, D. G., Gkioulidou, M., & Ukhorskiy, A. Y. (2018).
            Energetic ion injections inside geosynchronous orbit:
            Convection- and drift-dominated, charge-dependent adiabatic
            energization (W = qEd). Journal of Geophysical Research: Space
            Physics, 123, 6360–6382. https://doi.org/10.1029/2018JA025556

    .. [2]  Bingham, S. T., Cohen, I. J., Mauk, B. H., Turner, D. L.,
            Mitchell, D. G., Vines, S. K., et al. (2020).
            Charge-state‐dependent energization of suprathermal ions during
            substorm injections observed by MMS in the magnetotail. Journal
            of Geophysical Research: Space Physics, 125, e2020JA028144.
            https://doi.org/10.1029/2020JA028144

    """

    ne_s0 = len(flux_s0.energy.data)
    ne_s1 = len(flux_s1.energy.data)

    corr_mat = np.zeros((ne_s0, ne_s1))

    for i, j in itertools.product(range(ne_s0), range(ne_s1)):
        ref_flux = [flux_s0, flux_s1][np.argmax([len(flux_s0), len(flux_s1)])]
        flux_s0 = resample(flux_s0, ref_flux)
        flux_s1 = resample(flux_s1, ref_flux)

        # Clean time series
        current_flux_s0 = flux_s0.data[:, i]
        current_flux_s1 = flux_s1.data[:, j]

        # Replace zeros with nan
        current_flux_s0[current_flux_s0 == 0] = np.nan
        current_flux_s1[current_flux_s1 == 0] = np.nan

        # Remove NaNs
        cond = np.logical_not(np.logical_or(np.isnan(current_flux_s0),
                                            np.isnan(current_flux_s1)))

        # det_flux_s0 = signal.detrend(current_flux_s0[cond], type="constant")
        # det_flux_s1 = signal.detrend(current_flux_s1[cond], type="constant")

        det_flux_s0 = current_flux_s0[cond]
        det_flux_s1 = current_flux_s1[cond]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            corr_mat[i, j] = np.corrcoef(det_flux_s0, det_flux_s1)[0, 1]

    # Find largest block before NaNs
    idx_max_s0 = np.max(np.where(np.logical_not(np.isnan(corr_mat)))[0]) + 1
    idx_max_s1 = np.max(np.where(np.logical_not(np.isnan(corr_mat)))[1]) + 1

    energ_s0 = flux_s0.energy.data[:idx_max_s0]
    energ_s1 = flux_s1.energy.data[:idx_max_s1]
    corr_mat = corr_mat[:idx_max_s0, :idx_max_s1]

    return energ_s0, energ_s1, corr_mat



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz
from .calc_fs import calc_fs

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def convert_fac(inp, b_bgd, r_xyz: list = None):
    r"""Transforms to a field-aligned coordinate (FAC) system defined as :
        * R_parallel_z aligned with the background magnetic field
        * R_perp_y defined by R_parallel cross the position vector of the
            spacecraft (nominally eastward at the equator)
        * R_perp_x defined by R_perp_y cross R_par

    If inp is one vector along r direction, out is inp[perp, para]
    projection.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field.
    b_bgd : xarray.DataArray
        Time series of the background magnetic field.
    r_xyz : xarray.DataArray or ndarray or list
        Position vector of spacecraft.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input field in field aligned coordinates
        system.

    Notes
    -----
    All input parameters must be in the same coordinate system.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field (FGM) and electric field (EDP)

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Convert to field aligned coordinates

    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, numpy.array([1, 0, 0]))

    """

    assert r_xyz is None or isinstance(r_xyz, (xr.DataArray, list, np.ndarray))

    if r_xyz is None:
        r_xyz = np.array([1, 0, 0])

    if len(inp) != len(b_bgd):
        b_bgd = resample(b_bgd, inp, f_s=calc_fs(inp))

    time, inp_data = [inp.time.data, inp.data]

    # Normalize background magnetic field
    b_hat = b_bgd / np.linalg.norm(b_bgd, axis=1, keepdims=True)

    if isinstance(r_xyz, xr.DataArray):
        r_xyz = resample(r_xyz, b_bgd)

    elif len(r_xyz) == 3:
        r_xyz = np.tile(r_xyz, (len(b_bgd), 1))

    # Perpendicular
    r_perp_y = np.cross(b_hat, r_xyz, axis=1)
    r_perp_y /= np.linalg.norm(r_perp_y, axis=1, keepdims=True)
    r_perp_x = np.cross(r_perp_y, b_bgd, axis=1)
    r_perp_x /= np.linalg.norm(r_perp_x, axis=1, keepdims=True)

    assert inp_data.shape[1] in [1, 3], "Invalid dimension of inp"

    if inp_data.shape[1] == 3:
        out_data = np.zeros(inp.shape)

        out_data[:, 0] = np.sum(r_perp_x * inp_data, axis=1)
        out_data[:, 1] = np.sum(r_perp_y * inp_data, axis=1)
        out_data[:, 2] = np.sum(b_hat * inp_data, axis=1)

        # xarray
        out = xr.DataArray(out_data, coords=[time, inp.comp],
                           dims=["time", "comp"])

    else:
        out_data = np.zeros([3, inp_data.shape[0]])

        out_data[:, 0] = inp[:, 0] * np.sum(r_perp_x * r_xyz, axis=1)
        out_data[:, 1] = inp[:, 0] * np.sum(b_hat * r_xyz, axis=1)

        out = ts_vec_xyz(time, out_data, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import xarray as xr
import matplotlib.pyplot as plt

# Local imports
from ..pyrf import histogram2d
from . import plot_spectr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def pl_scatter_matrix(inp1, inp2: xr.DataArray = None, pdf: bool = False,
                      cmap: str = "jet"):
    r"""Produces a scatter plot of each components of field inp1 with respect
    to every component of field inp2. If pdf is set to True, the scatter
    plot becomes a 2d histogram.

    Parameters
    ----------
    inp1 : xarray.DataArray
        First time series (x-axis).
    inp2 : xarray.DataArray
        Second time series (y-axis).
    pdf : bool, Optional
        Flag to plot the 2d histogram. If False the figure is a scatter plot.
        If True the figure is a 2d histogram.
    cmap : str, Optional
        Colormap. Default : "jet"

    Returns
    -------
    fig : matplotlib.pyplot.figure
        Figure with time series plots.
    axs : matplotlib.axes._subplots.AxesSubplot
        Axes.
    caxs : matplotlib.pyplot.colorbar
        Only if pdf is True

    """

    if inp2 is None:
        inp2 = inp1
        warnings.warn("inp2 is empty assuming that inp2=inp1", UserWarning)

    assert isinstance(inp1, xr.DataArray) and isinstance(inp2, xr.DataArray)

    if not pdf:
        fig, axs = plt.subplots(3, 3, sharex="all", sharey="all",
                                figsize=(16, 9))
        fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9,
                            hspace=0.05, wspace=0.05)

        for i in range(3):
            for j in range(3):
                axs[j, i].scatter(inp1[:, i].data, inp2[:, j].data, marker="+")

        out = (fig, axs)
    else:
        fig, axs = plt.subplots(3, 3, sharex="all", sharey="all",
                                figsize=(16, 9))
        fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9,
                            hspace=0.05, wspace=0.3)

        caxs = [[None]*3]*3

        for i in range(3):
            for j in range(3):
                hist_ = histogram2d(inp1[:, i], inp2[:, j])
                axs[j, i], caxs[j][i] = plot_spectr(axs[j, i], hist_,
                                                    cmap=cmap, cscale="log")
                axs[j, i].grid()

        out = (fig, axs, caxs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _phxtof_calibration(energy, alpha, beta, gamma):
    r"""Pulse Height x Time Of Flight correction model from EPD Data Product
    Guide"""
    return 1 / (.5 * (1 + alpha * (np.tanh((energy - beta) / gamma) + 1)))


def _extof_calibration(energy, alpha, beta, gamma):
    r"""Energy x Time Of Flight correction model from EPD Data Product Guide"""
    return 1 / (.5 * (1 + alpha * (1 - np.tanh((energy - beta) / gamma) + 1)))


def eis_proton_correction(flux_eis):
    r"""Corrects proton flux values based on FPI/HPCA/EPD-EIS cross
    calibration.

    Parameters
    ----------
    flux_eis : xarray.DataArray
        Omni-directional energy spectrum from EPD-EIS.

    Returns
    -------
    flux_eis_corr : xarray.DataArray
        Cross-calibrated omni-directional energy spectrum from EIS-EPD.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_omni

    """

    #  Coefficients from EPD Data Product Guide
    alpha_, beta_, gamma_ = [-.3, 49e-3, 1e-3]

    # Pulse Height x Time Of Flight (PHxTOF) energy correction factor
    energy_phxtof = flux_eis.energy.data[:7]
    phxtof_corr = _phxtof_calibration(energy_phxtof, alpha_, beta_, gamma_)

    # Energy x Time Of Flight (ExTOF) energy correction factor
    energy_extof = flux_eis.energy.data[7:]
    extof_corr = _extof_calibration(energy_extof, alpha_, beta_, gamma_)

    eis_corr = np.hstack([phxtof_corr, extof_corr])

    if isinstance(flux_eis, xr.Dataset):
        scopes_eis = list(filter(lambda x: x[0] == "t", flux_eis))
        out_keys = list(filter(lambda x: x not in scopes_eis, flux_eis))
        out_dict = {k: flux_eis[k] for k in out_keys}

        for scope in scopes_eis:
            out_dict[scope] = flux_eis[scope].copy()
            out_dict[scope].data *= eis_corr

        flux_eis_corr = xr.Dataset(out_dict)
    elif isinstance(flux_eis, xr.DataArray):
        # Apply correction to omni-directional energy spectrum
        flux_eis_corr = flux_eis.copy()
        flux_eis_corr.data *= eis_corr
    else:
        raise TypeError("flux_eis must be a Dataset or a DataArray")

    return flux_eis_corr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os
import re
import bisect
import datetime

# 3rd party imports
from dateutil import parser
from dateutil.rrule import rrule, DAILY

# Local imports
from .mms_config import CONFIG

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def list_files(tint, mms_id, var, data_path=""):
    """Find files in the data directories of the target instrument, data type,
    data rate, mms_id and level during the target time interval.

    Parameters
    ----------
    tint : list
        Time interval
    mms_id : str or int
        Index of the spacecraft
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`

    Returns
    -------
    files : list
        List of files corresponding to the parameters in the selected time
        interval

    """

    # Check path
    if not data_path:
        data_path = CONFIG["local_data_dir"]

    files_out = []

    if not isinstance(mms_id, str):
        mms_id = str(mms_id)
    # directory and file name search patterns:
    # - assume directories are of the form:
    # (srvy, SITL): spacecraft/instrument/rate/level[/datatype]/year/month/
    # (brst): spacecraft/instrument/rate/level[/datatype]/year/month/day/
    # - assume file names are of the form:
    # spacecraft_instrument_rate_level[_datatype]_YYYYMMDD[hhmmss]_version.cdf

    file_name = f"mms{mms_id}_{var['inst']}_{var['tmmode']}_{var['lev']}" \
                + r"(_)?.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    d_start = parser.parse(parser.parse(tint[0]).strftime("%Y-%m-%d"))
    until_ = parser.parse(tint[1]) - datetime.timedelta(seconds=1)
    days = rrule(DAILY, dtstart=d_start, until=until_)

    if var["dtype"] == "" or var["dtype"] is None:
        level_and_dtype = var["lev"]
    else:
        level_and_dtype = os.sep.join([var["lev"], var["dtype"]])

    for date in days:
        if var["tmmode"] == "brst":
            local_dir = os.sep.join([data_path, f"mms{mms_id}", var["inst"],
                                     var["tmmode"], level_and_dtype,
                                     date.strftime("%Y"), date.strftime("%m"),
                                     date.strftime("%d")])
        else:
            local_dir = os.sep.join([data_path, f"mms{mms_id}", var["inst"],
                                     var["tmmode"], level_and_dtype,
                                     date.strftime("%Y"), date.strftime("%m")])

        if os.name == "nt":
            full_path = os.sep.join([re.escape(local_dir)+os.sep, file_name])
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        regex = re.compile(full_path)

        for root, dirs, files in os.walk(local_dir):
            for file in files:
                this_file = os.sep.join([root, file])

                matches = regex.match(this_file)
                if matches:
                    this_time = parser.parse(matches.groups()[1])
                    if d_start <= this_time <= until_:
                        if this_file not in files_out:
                            files_out.append({"file_name": file,
                                              "timetag": "",
                                              "full_name": this_file,
                                              "file_size": ""})

    in_files = files_out

    file_name = r"mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    file_times = []

    regex = re.compile(file_name)

    for file in in_files:
        matches = regex.match(file["file_name"])
        if matches:
            file_times.append((file["file_name"],
                               parser.parse(matches.groups()[0]).timestamp(),
                               file["timetag"], file["file_size"]))

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect.bisect_left(times, parser.parse(tint[0]).timestamp())

    # note: purposefully liberal here; include one extra file so that we
    # always get the burst mode
    # data
    if idx_min == 0:
        files_in_interval = []
        for file in sorted_files[idx_min:]:
            files_in_interval.append({"file_name": file[0],
                                      "timetag": file[2],
                                      "file_size": file[3]})
    else:
        files_in_interval = []
        for file in sorted_files[idx_min - 1:]:
            files_in_interval.append({"file_name": file[0],
                                      "timetag": file[2],
                                      "file_size": file[3]})

    local_files = []

    file_names = [f["file_name"] for f in files_in_interval]

    for file in files_out:
        if file["file_name"] in file_names:
            local_files.append(file["full_name"])

    return sorted(local_files)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os

# 3rd party imports
import numpy as np
import xarray as xr
import numba

from scipy import fft

# Local imports
from .calc_fs import calc_fs

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _scales(f_range, f_nyq, f_s, n_freqs, linear_df, delta_f):
    if linear_df:
        scale_number = np.floor(f_nyq / delta_f).astype(int)

        f_range = [delta_f, scale_number * delta_f]

        scales = f_nyq / (np.linspace(f_range[0], f_range[1], scale_number))
    else:
        scale_min = np.log10(.5 * f_s / f_range[1])
        scale_max = np.log10(.5 * f_s / f_range[0])
        scales = np.logspace(scale_min, scale_max, n_freqs)

    return scales


@numba.jit(nopython=True, fastmath=True)
def _ww(s_ww, scales_mat, sigma, frequencies_mat, f_nyq):
    w_w = s_ww * np.exp(-sigma * sigma
                        * ((scales_mat * frequencies_mat - f_nyq) ** 2) / 2)
    w_w = w_w * np.sqrt(1)
    return w_w


@numba.jit(nopython=True, parallel=True, fastmath=True)
def _power_r(power, new_freq_mat):
    power2 = np.absolute((2 * np.pi) * np.conj(power) * power / new_freq_mat)
    return power2


@numba.jit(nopython=True, parallel=True, fastmath=True)
def _power_c(power, new_freq_mat):
    power2 = np.sqrt(np.absolute((2 * np.pi) / new_freq_mat)) * power
    return power2


def wavelet(inp, f_range: list = None, n_freqs: int = 200,
            w_width: float = 5.36, delta_f: float = 100.,
            linear_df: bool = False, cut_edge: bool = True,
            return_power: bool = True):
    r"""Computes wavelet spectrogram based on fast FFT algorithm.

    Parameters
    ----------
    inp : xarray.DataArray
        Input quantity.
    f_range : list, Optional
        Vector [f_min f_max], calculate spectra between frequencies f_min
        and f_max.
    n_freqs : int, Optional
        Number of frequency bins. Default is 200.
    w_width : float, Optional
        Width of the Morlet wavelet. Default 5.36.
    delta_f : float, Optional
        Spacing between frequencies. Default is 100.
    linear_df : bool, Optional
        Linear spacing between frequencies of df. Default is False (use
        log spacing)
    return_power : bool, Optional
        Set to True to return the power, False for complex wavelet
        transform. Default True.
    cut_edge : bool, Optional
        Set to True to set points affected by edge effects to NaN, False to
        keep edge affect points. Default True.

    Returns
    -------
    out : xarray.DataArray or xarray.Dataset
        Wavelet transform of the input.

    """

    # Unpack time and data
    data = inp.data

    # Compute sampling frequency
    f_s = calc_fs(inp)

    scale_min, scale_max = [0.01, 2]

    if f_range is None:
        f_range = [.5 * f_s / 10 ** scale_max, .5 * f_s / 10 ** scale_min]

    f_nyq, scale_number, sigma = [f_s / 2, n_freqs, w_width / (f_s / 2)]

    scales = _scales(f_range, f_nyq, f_s, n_freqs, linear_df, delta_f)

    # Remove the last sample if the total number of samples is odd
    if len(data) / 2 != np.floor(len(data) / 2):
        time = inp.time.data[:-1]
        data = data[:-1, ...]
    else:
        time = inp.time.data

    # Check for NaNs
    scales[np.isnan(scales)] = 0

    # Find the frequencies for an FFT of all data
    freq = f_s * .5 * np.arange(1, 1 + len(data) / 2) / (len(data) / 2)

    # The frequencies corresponding to FFT
    frequencies = np.hstack([0, freq, -np.flip(freq[:-1])])

    # Get the correct frequencies for the wavelet transform
    new_freq = f_nyq / scales

    new_freq_mat, temp_freq = np.meshgrid(new_freq, frequencies, sparse=True)

    _, frequencies_mat = np.meshgrid(scales, frequencies, sparse=True)

    if len(inp.shape) == 1:
        data = data[:, np.newaxis]  # if scalar add virtual axis
        out_dict, power2 = [None, np.zeros((len(inp.data), n_freqs))]
    elif len(inp.shape) == 2:
        out_dict, power2 = [{}, None]
    else:
        raise TypeError("Invalid shape of the inp")

    # go through all the data columns
    for i in range(data.shape[1]):
        # Make the FFT of all data
        data_col = data[:, i]

        # Wavelet transform of the data
        # Forward FFT
        s_w = fft.fft(data_col, workers=os.cpu_count())

        scales_mat, s_w_mat = np.meshgrid(scales, s_w, sparse=True)

        # Calculate the FFT of the wavelet transform
        w_w = _ww(s_w_mat, scales_mat, sigma, frequencies_mat, f_nyq)

        # Backward FFT
        power = fft.ifft(w_w, axis=0, workers=os.cpu_count())

        # Calculate the power spectrum
        if return_power:
            power2 = _power_r(power, np.tile(new_freq_mat, (len(power), 1)))
        else:
            power2 = _power_c(power, np.tile(new_freq_mat, (len(power), 1)))

        if cut_edge:
            censure = np.floor(2 * scales).astype(int)

            for j in range(scale_number):
                power2[:censure[j], j] = np.nan

                power2[len(data_col) - censure[j]:len(data_col), j] = np.nan
        if len(inp.shape) == 2:
            out_dict[inp.comp.data[i]] = (["time", "frequency"], power2)

    if len(inp.shape) == 1:
        out = xr.DataArray(power2, coords=[time, new_freq],
                           dims=["time", "frequency"])
    elif len(inp.shape) == 2:
        out = xr.Dataset(out_dict,
                         coords={"time": time, "frequency": new_freq})
    else:
        raise TypeError("Invalid shape")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"

plt.style.use("seaborn-ticks")
colors_ = ["tab:blue", "tab:green", "tab:red", "k"]
plt.rc('axes', prop_cycle=mpl.cycler(color=colors_))


def pl_tx(axis, inp_list, comp, **kwargs):
    r"""Line plot of 4 spacecraft time series.

    Parameters
    ----------
    axis : matplotlib.axes._subplots.AxesSubplot
        Axis
    inp_list : list of xarray.DataArray
        Time series to plot
    comp: int
        Index of the column to plot.

    Other Parameters
    ----------------
    kwargs : dict
        Hash table of plot options.

    """

    if axis is None:
        _, axis = plt.subplots(1)

    for inp in inp_list:
        if len(inp.shape) == 3:
            data = np.reshape(inp.data,
                              (inp.shape[0], inp.shape[1] * inp.shape[2]))
        elif len(inp.shape) == 1:
            data = inp.data[:, np.newaxis]
        else:
            data = inp.data

        time = inp.time
        axis.plot(time, data[:, comp], **kwargs)

    locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
    formatter = mdates.ConciseDateFormatter(locator)
    axis.xaxis.set_major_locator(locator)
    axis.xaxis.set_major_formatter(formatter)
    axis.grid(True, which="major", linestyle="-", linewidth="0.5", c="0.5")

    return axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eis_omni(eis_allt):
    r"""Calculates the omni-directional flux for all 6 telescopes.

    Parameters
    ----------
    eis_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.

    Returns
    -------
    flux_omni : xarray.DataArray
        Omni-directional flux for all 6 telescopes

    See Also
    --------
    pyrfu.mms.get_eis_allt

    Examples
    --------
    >>> from pyrfu import mms

    Define spacecraft index and time interval

    >>> tint = ["2017-07-23T16:10:00", "2017-07-23T18:10:00"]
    >>> ic = 2

    Get EIS ExTOF all 6 telescopes fluxes

    >>> extof_allt = mms.get_eis_allt("flux_extof_proton_srvy_l2", tint, ic)

    Compute the omni-directional flux for all 6 telescopes

    >>> extof_omni = mms.eis_omni(extof_allt)

    """

    scopes = list(filter(lambda x: x[0] == "t", eis_allt))

    flux_omni = None

    for scope in scopes:
        try:
            flux_omni += eis_allt[scope].copy()
        except TypeError:
            flux_omni = eis_allt[scope].copy()

    flux_omni.data /= len(scopes)
    flux_omni.name = "flux_omni"
    flux_omni.attrs["energy_dplus"] = eis_allt.energy_dplus.data
    flux_omni.attrs["energy_dminus"] = eis_allt.energy_dminus.data

    return flux_omni



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


all_params_scalars = ["ni", "nbgi", "pbgi", "partni", "ne", "pbge", "nbge",
                      "partne", "tsi", "tperpi", "tparai", "parttperpi",
                      "parttparai", "tse", "tperpe", "tparae", "parttperpe",
                      "parttparae", "pde", "pdi", "pderre", "pderri", "v",
                      "v6", "defi", "defbgi", "defe", "defbge", "energyi",
                      "bnergye", "epar", "sdev12", "sdev34", "flux-amb-pm2",
                      "padlowene", "padmidene", "padhighene", "bpsd", "epsd"]

all_params_vectors = ["r", "sti", "vi", "errvi", "partvi", "ste", "ve",
                      "errve", "partve", "qi", "errqi", "qe", "errqe", "b",
                      "e", "e2d", "es12", "es34"]

all_params_tensors = ["pi", "partpi", "pe", "partpe", "ti", "partti", "te",
                      "partte"]

hpca_params_scalars = ["nhplus", "nheplus", "nheplusplus", "noplus",
                       "tshplus", "tsheplus", "tsheplusplus", "tsoplus",
                       "dpfhplus", "dpfheplus", "dpfheplusplus",
                       "dpfoplus", "phase", "adcoff", "saz", "azimuth"]

hpca_params_tensors = ["vhplus", "vheplus", "vheplusplus", "voplus",
                       "phplus", "pheplus", "pheplusplus", "poplus",
                       "thplus", "theplus", "theplusplus", "toplus"]

instruments = ["mec", "fpi", "edp", "edi", "hpca", "fgm",
               "dfg", "afg", "scm", "fsm", "dsp"]

coordinate_systems = ["gse", "gsm", "dsl", "dbcs", "dmpa", "ssc", "bcs", "par"]

data_lvls = ["ql", "sitl", "l1b", "l2a", "l2pre", "l2", "l3"]


def _tensor_order(splitted_key):
    param_low = splitted_key[0].lower()

    if param_low in all_params_scalars or param_low in hpca_params_scalars:
        tensor_order = 0
    elif param_low in all_params_vectors or param_low in hpca_params_tensors:
        tensor_order = 1
    elif param_low in all_params_tensors:
        tensor_order = 2
    else:
        raise ValueError(f"invalid PARAM : {param_low}")

    return tensor_order


def tokenize(var_str):
    r"""Parses the variable keys.

    Parameters
    ----------
    var_str : str
        Input key of variable

    Returns
    -------
    out : dict
        Hash table containing :
            * param : Variable key.
            * to : Tensor order.
            * cs : Coordinate system.
            * inst : Instrument.
            * tmmode : Time mode.
            * lev" : Level of data.

    """

    splitted_key = var_str.split("_")

    tensor_order = _tensor_order(splitted_key)

    coordinate_system = []
    idx = 0

    if tensor_order > 0:
        coordinate_system = splitted_key[idx + 1]
        assert coordinate_system in coordinate_systems, "invalid coord. sys."
        idx += 1

    instrument = splitted_key[idx + 1]
    assert instrument in instruments, "invalid INSTRUMENT"

    idx += 1

    tmmode = splitted_key[idx + 1]
    idx += 1

    if tmmode not in ["brst", "fast", "slow", "srvy"]:
        tmmode = "fast"
        idx -= 1
        warnings.warn("assuming TM_MODE = FAST", UserWarning)

    if len(splitted_key) == idx + 1:
        data_lvl = "l2"  # default
    else:
        data_lvl = splitted_key[idx + 1]
        assert data_lvl in data_lvls, "invalid DATA_LEVEL level"

    res = {"param": splitted_key[0], "to": tensor_order,
           "cs": coordinate_system, "inst": instrument,
           "tmmode": tmmode, "lev": data_lvl}

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os
import re
import json
import glob
import bisect
import fnmatch
import logging

# 3rd party imports
import pandas as pd

# Local imports
from ..pyrf import iso86012datetime

from .mms_config import CONFIG

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S',
                    level=logging.INFO)


def load_ancillary(level_and_dtype, tint, mms_id, verbose: bool = True,
                   data_path: str = ""):
    r"""Loads ancillary data.

    Parameters
    ----------
    level_and_dtype : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.
    tint : list of str
        Time interval
    mms_id : str or int
        Spacecraft index
    verbose : bool, Optional
        Set to True to follow the loading. Default is True
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu.mms.mms_config.py`

    Returns
    -------
    out : xarray.Dataset
        Time series of the ancillary data

    """

    if not data_path:
        data_path = CONFIG["local_data_dir"]

    if isinstance(mms_id, int):
        mms_id = str(mms_id)

    tint = iso86012datetime(tint)

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    dir_pattern = os.sep.join([data_path, "ancillary", f"mms{mms_id}",
                               level_and_dtype])
    file_pattern = "_".join(["MMS{}".format(mms_id), level_and_dtype.upper(),
                             "???????_???????.V??"])

    files_in_tint = []
    out_files = []

    files = glob.glob(os.sep.join([dir_pattern, file_pattern]))

    # find the files within the time interval
    fname_fmt = f"MMS{mms_id}_{level_and_dtype.upper()}" \
                f"_([0-9]{{7}})_([0-9]{{7}}).V[0-9]{{2}}"
    file_regex = re.compile(os.sep.join([dir_pattern, fname_fmt]))
    for file in files:
        time_match = file_regex.match(file)
        if time_match is not None:
            start_time = pd.to_datetime(time_match.group(1), format="%Y%j")
            end_time = pd.to_datetime(time_match.group(2), format="%Y%j")
            if start_time < tint[1] and end_time >= tint[0]:
                files_in_tint.append(file)

    # ensure only the latest version of each file is loaded
    for file in files_in_tint:
        this_file = file[0:-3] + "V??"
        versions = fnmatch.filter(files_in_tint, this_file)
        if len(versions) > 1:
            # only grab the latest version
            out_files.append(sorted(versions)[-1])
        else:
            out_files.append(versions[0])

    files_names = list(set(out_files))
    files_names.sort()

    # Read length of header and columns names from .json file
    # Root path
    root_path = os.path.dirname(os.path.abspath(__file__))

    with open(os.sep.join([root_path, "ancillary.json"])) as file:
        anc_dict = json.load(file)

    if verbose:
        logging.info(f"Loading ancillary {level_and_dtype} files...")

    data_frame_dict = {}

    for i, file in enumerate(files_names):
        rows = pd.read_csv(file, delim_whitespace=True, header=None,
                           skiprows=anc_dict[level_and_dtype]["header"])

        # Remove footer
        rows = rows[:][:-1]

        # Convert time
        fmt = anc_dict[level_and_dtype]["time_format"]
        rows[0] = pd.to_datetime(rows[0], format=fmt)

        start_idx = bisect.bisect_left(rows[0][:], tint[0])
        end_idx = bisect.bisect_left(rows[0][:], tint[1])
        rows.columns = anc_dict[level_and_dtype]["columns_names"]

        data_frame_dict[i] = rows[:][start_idx:end_idx]

    data_frame = data_frame_dict[0]

    for k in list(data_frame_dict.keys())[1:]:
        data_frame = data_frame.append(data_frame_dict[k])

    data_frame = data_frame.sort_values(by="time").set_index(["time"])

    return data_frame.to_xarray()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

# Local imports
from .list_files import list_files
from .db_get_ts import db_get_ts
from .db_get_variable import db_get_variable

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def get_eis_allt(tar_var, tint, mms_id, verbose: bool = True,
                 data_path: str = ""):
    r"""Read energy spectrum of the selected specie in the selected energy
    range for all telescopes.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like
        {data_unit}_{dtype}_{specie}_{data_rate}_{data_lvl}.
    tint : list of str
        Time interval.
    mms_id : int or float or str
        Index of the spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Path of MMS data.

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the 6 telescopes of the
        Energy Ion Spectrometer.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2017-07-23T16:54:24.000", "2017-07-23T17:00:00.000"]

    Read proton energy spectrum for all EIS telescopes

    >>> eis_allt = mms.get_eis_allt("Flux_extof_proton_srvy_l2", tint_brst, 2)

    """

    # Convert mms_id to integer
    mms_id = int(mms_id)

    data_unit, data_type, specie, data_rate, data_lvl = tar_var.split("_")

    pref = f"mms{mms_id:d}_epd_eis"

    var = {"mms_id": mms_id, "inst": "epd-eis", "dtype": data_type,
           "tmmode": data_rate, "lev": data_lvl, "specie": specie,
           "data_path": data_path}

    # EIS includes the version of the files in the cdfname need to read it
    # before.
    files = list_files(tint, mms_id, var, data_path=data_path)

    file_version = int(files[0].split("_")[-1][1])
    var["version"] = file_version

    if int(files[0].split("_")[-1][3]) >= 1:
        pref = f"{pref}_{data_rate}_{data_lvl}_{data_type}"
        if specie == "alpha":
            specie = "helium"
    else:
        if data_rate == "brst":
            pref = f"{pref}_{data_rate}_{data_type}"
        else:
            pref = f"{pref}_{data_type}"

    if data_unit.lower() in ["flux", "counts", "cps"]:
        suf = f"{specie}_P{file_version:d}_{data_unit.lower()}_t"
    else:
        raise ValueError("Invalid data unit")

    # Name of the data containing index of the probe, instrument, data rate,
    # data level and data type if needed
    dset_name = f"mms{var['mms_id']:d}_{var['inst']}_{var['tmmode']}" \
                f"_{var['lev']}_{var['dtype']}"

    # Names of the energy spectra in the CDF (one for each telescope)
    cdfnames = ["{}_{}{:d}".format(pref, suf, t) for t in range(6)]

    spin_nums = db_get_ts(dset_name, f"{pref}_spin", tint, data_path=data_path)
    sectors = db_get_ts(dset_name, f"{pref}_sector", tint, data_path=data_path)

    e_minu = db_get_variable(dset_name, f"{pref}_{specie}_t0_energy_dminus",
                             tint, verbose=verbose, data_path=data_path)

    e_plus = db_get_variable(dset_name, f"{pref}_{specie}_t0_energy_dplus",
                             tint, verbose=verbose, data_path=data_path)

    outdict = {"spin": spin_nums, "sector": sectors}

    for i, cdfname in enumerate(cdfnames):
        scope_key = f"t{i:d}"

        outdict[scope_key] = db_get_ts(dset_name, cdfname, tint,
                                       verbose=verbose, data_path=data_path)
        outdict[scope_key] = outdict[scope_key].rename({"time": "time",
                                                        "Energy": "energy"})

        outdict[f"look_{scope_key}"] = db_get_ts(dset_name,
                                                 f"{pref}_look_{scope_key}",
                                                 tint, verbose=verbose,
                                                 data_path=data_path)

    e_plus = e_plus.assign_coords(x=outdict["t0"].energy.data)
    e_minu = e_minu.assign_coords(x=outdict["t0"].energy.data)
    e_plus = e_plus.rename({"x": "energy"})
    e_minu = e_minu.rename({"x": "energy"})

    outdict["energy_dplus"] = e_plus
    outdict["energy_dminus"] = e_minu

    # Build Dataset
    out = xr.Dataset(outdict, attrs=var)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import pandas as pd

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def datetime_to_tt2000(time):
    r"""Transforms datetime to TT2000 string format.

    Parameters
    ----------
    time : datetime.datetime
        Time to convert to tt2000 string.

    Returns
    -------
    tt2000 : str
        Time in TT20000 iso_8601 format.

    """

    time_datetime = pd.Timestamp(time)

    # Convert to string
    datetime_str = time_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f')

    tt2000 = f"{datetime_str}{time_datetime.nanosecond:03d}"

    return tt2000



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf import ts_skymap

from .dpf2psd import dpf2psd

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def eis_skymap(inp_ang_ang, to_psd: bool = True):
    r"""Construct skymap distribution from angle-angle distribution.

    Parameters
    ----------
    inp_ang_ang : xarray.DataArray
        EIS angle-angle distribution.
    to_psd : bool, Optional
        Flag to convert differential particle flux to phase space density.

    Returns
    -------
    out : xarray.Dataset
        EIS skymap distribution.

    See Also
    --------
    pyrfu.mms.eis_ang_ang

    """

    time_data = inp_ang_ang.time.data
    enr_ = np.tile(inp_ang_ang.energy.data, (len(time_data), 1))
    phi_ = np.tile(inp_ang_ang.phi.data, (len(time_data), 1))
    the_ = inp_ang_ang.theta.data

    out = ts_skymap(time_data, inp_ang_ang.data, enr_, phi_, the_)
    out.attrs["species"] = "ions"
    out.attrs["UNITS"] = "1/(cm^2 s sr keV)"

    out.energy.data *= 1e3

    if to_psd:
        out = dpf2psd(out)

    out.attrs["energy_dminus"] = inp_ang_ang.attrs["energy_dminus"]
    out.attrs["energy_dplus"] = inp_ang_ang.attrs["energy_dplus"]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def _check_spin(spin):
    _, spin_inds = np.unique(spin, return_index=True)

    # Check to see if we have a complete first spin,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[0]])[0].size != 8:
        spin_inds = spin_inds[1:]

    # Check to see if the last one is complete,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[-1]])[0].size != 8:
        spin_inds = spin_inds[:-1]

    return spin_inds


def eis_ang_ang(inp_allt, en_chan: list = None):
    r"""Generates EIS angle-angle distribution.

    Parameters
    ----------
    inp_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.
    en_chan : array_like, Optional
        Energy channels to use. Default use all energy channels.

    Returns
    -------
    out : xarray.DataArray
        EIS skymap-like distribution.

    """

    if en_chan is None:
        en_chan = np.arange(len(inp_allt.energy))

    time_ = inp_allt.time.data

    n_en = len(en_chan)

    scopes = list(filter(lambda x: x.startswith("t"), inp_allt.keys()))

    azi_ = np.zeros((len(scopes), len(inp_allt.time)))
    pol_ = np.zeros((len(scopes), len(inp_allt.time)))

    for i, scope in enumerate(scopes):
        d_ = inp_allt[f"look_{scope}"]
        # Domain [-180,180], 0 = sunward (GSE)
        azi_[i, :] = np.rad2deg(np.arctan2(d_.data[:, 1], d_.data[:, 0]))
        # Domain [-90,90], Positive is look direction northward
        pol_[i, :] = 90. - np.rad2deg(np.arccos(d_[:, 2]))

    spin_ = inp_allt.spin.data
    sect_ = inp_allt.sector.data

    spin_inds = _check_spin(spin_)

    n_spins, n_pol, n_azi = [len(spin_inds), 6, len(np.unique(sect_))]

    # Minus 80 plus
    min_pol_edges = -80. + 160. * np.arange(n_pol) / n_pol
    max_pol_edges = -80. + 160. * (np.arange(n_pol) + 1) / n_pol

    min_azi_edges = -180. + 360. * np.arange(n_azi) / n_azi
    max_azi_edges = -180. + 360. * (np.arange(n_azi) + 1) / n_azi

    out_data = np.zeros((n_spins, n_en, n_azi, n_pol))

    time_data = time_[spin_inds]

    for i, spin_ind in enumerate(spin_inds):
        t_inds = np.where(spin_ == spin_[spin_ind])[0]
        for t_ind, (t, scope) in itertools.product(t_inds, enumerate(scopes)):
            cond_azi = np.logical_and(azi_[t, t_ind] > min_azi_edges,
                                      azi_[t, t_ind] < max_azi_edges)
            a_idx = np.where(cond_azi)[0]

            cond_pol = np.logical_and(pol_[t, t_ind] > min_pol_edges,
                                      pol_[t, t_ind] < max_pol_edges)
            p_idx = np.where(cond_pol)[0]

            out_data[i, :, a_idx, p_idx] = inp_allt[scope].data[t_ind, en_chan]

    out = xr.DataArray(out_data,
                       coords=[time_data, inp_allt.energy.data[en_chan],
                               min_azi_edges + 180. / n_azi,
                               min_pol_edges + 90. / n_pol],
                       dims=["time", "energy", "phi", "theta"])

    out.attrs["energy_dminus"] = inp_allt.energy_dminus.data
    out.attrs["energy_dplus"] = inp_allt.energy_dplus.data

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# @Louis Richard
from .ts_time import ts_time
from .ts_scalar import ts_scalar
from .ts_vec_xyz import ts_vec_xyz
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_skymap import ts_skymap
from .ts_append import ts_append
from .dist_append import dist_append
from .start import start
from .end import end
from .iso2unix import iso2unix
from .extend_tint import extend_tint
from .dot import dot
from .cross import cross
from .trace import trace
from .norm import norm
from .normalize import normalize
from .calc_dt import calc_dt
from .calc_fs import calc_fs
from .avg_4sc import avg_4sc
from .gradient import gradient
from .integrate import integrate
from .time_clip import time_clip
from .resample import resample
from .t_eval import t_eval
from .filt import filt
from .medfilt import medfilt
from .movmean import movmean
from .wavelet import wavelet
from .compress_cwt import compress_cwt
from .mva import mva
from .vht import vht
from .new_xyz import new_xyz
from .dec_par_perp import dec_par_perp
from .convert_fac import convert_fac
from .edb import edb
from .e_vxb import e_vxb
from .plasma_calc import plasma_calc
from .dynamic_press import dynamic_press
from .pres_anis import pres_anis
from .calc_sqrtq import calc_sqrtq
from .median_bins import median_bins
from .mean_bins import mean_bins
from .histogram import histogram
from .histogram2d import histogram2d
from .wavepolarize_means import wavepolarize_means
from .psd import *
from .c_4_k import c_4_k
from .c_4_grad import c_4_grad
from .c_4_j import c_4_j
from .c_4_v import c_4_v
from .ebsp import ebsp
from .date_str import date_str
from .find_closest import find_closest
from .cart2sph_ts import cart2sph_ts
from .corr_deriv import corr_deriv
from .eb_nrf import eb_nrf
from .lowpass import lowpass
from .mean import mean
from .poynting_flux import poynting_flux
from .remove_repeated_points import remove_repeated_points
from .solid_angle import solid_angle
from .wave_fft import wave_fft
from .pvi import pvi
from .pvi_4sc import pvi_4sc
from .pid_4sc import pid_4sc
from .calc_dng import calc_dng
from .calc_ag import calc_ag
from .calc_agyro import calc_agyro
from .datetime_to_tt2000 import datetime_to_tt2000
from .cotrans import cotrans
from .sph2cart import sph2cart
from .gse2gsm import gse2gsm
from .mf import mf
from .waverage import waverage
from .l_shell import l_shell
from .optimize_nbins_1d import optimize_nbins_1d
from .optimize_nbins_2d import optimize_nbins_2d
from .magnetosphere import magnetosphere
from .iso86012timevec import iso86012timevec
from .timevec2iso8601 import timevec2iso8601
from .iso86012datetime64 import iso86012datetime64
from .datetime642iso8601 import datetime642iso8601
from .unix2datetime64 import unix2datetime64
from .datetime642unix import datetime642unix
from .ttns2datetime64 import ttns2datetime64
from .datetime642ttns import datetime642ttns
from .iso86012datetime import iso86012datetime
from .cart2sph import cart2sph
from .iplasma_calc import iplasma_calc
from .match_phibe_dir import match_phibe_dir
from .match_phibe_v import match_phibe_v
from .get_omni_data import get_omni_data
from .estimate import estimate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pdb

# 3rd party imports
import numpy as np

from scipy import constants

# Local imports
from ..pyrf import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def remove_idist_background(n_i, v_gse_i, p_gse_i, n_bg_i, p_bg_i):
    r"""Removes penetrating radiation background from ion moments.

    Parameters
    ----------
    n_i : xarray.DataArray
        Time series of the ion density.
    v_gse_i : xarray.DataArray
        Time series of the ion bulk velocity.
    p_gse_i : xarray.DataArray
        Time series of the ion pressure tensor.
    n_bg_i : xarray.DataArray
        Time series of the background ion number density.
    p_bg_i : xarray.DataArray
        Time series of the background ion pressure scalar.

    Returns
    -------
    n_i_new : xarray.DataArray
        Time series of the corrected ion number density.
    v_gse_i_new : xarray.DataArray
        Time series of the corrected ion bulk velocity.
    p_gse_i : xarray.DataArray
        Time series of the corrected ion pressure tensor.

    References
    ----------
    MMS DIS Penetrating radiation correction methods.

    """

    m_p = constants.proton_mass

    # Number density
    n_i_new = n_i - n_bg_i.data

    # Bulk velocity
    v_gse_i_new = v_gse_i.copy()
    v_gse_i_new.data *= n_i.data[:, None] / n_i_new.data[:, None]

    # Pressure tensor
    p_gse_i_new = np.zeros(p_gse_i.shape)
    n_old, v_old = [n_i.data, v_gse_i.data]
    n_new, v_new = [n_i_new.data, v_gse_i_new.data]

    for i, j in zip([0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]):
        p_gse_i_new[:, i, j] += p_gse_i.data[:, 0, 0]
        p_gse_i_new[:, i, j] -= p_bg_i.data
        p_gse_i_new[:, i, j] += m_p * n_old * np.multiply(v_old[:, i],
                                                          v_old[:, j])
        p_gse_i_new[:, i, j] -= m_p * n_new * np.multiply(v_new[:, i],
                                                          v_new[:, j])

    # Pressure tensor is symmetric
    p_gse_i_new[:, 1, 0] = p_gse_i_new[:, 0, 1]
    p_gse_i_new[:, 2, 0] = p_gse_i_new[:, 0, 2]
    p_gse_i_new[:, 2, 1] = p_gse_i_new[:, 1, 2]

    p_gse_i_new = ts_tensor_xyz(p_gse_i.time.data, p_gse_i_new)

    return n_i_new, v_gse_i_new, p_gse_i_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# @Louis Richard
from .tokenize import tokenize
from .list_files import list_files
from .get_ts import get_ts
from .get_dist import get_dist
from .db_get_ts import db_get_ts
from .fk_power_spectrum_4sc import fk_power_spectrum_4sc
from .lh_wave_analysis import lh_wave_analysis
from .whistler_b2e import whistler_b2e
from .get_feeps_omni import get_feeps_omni
from .remove_idist_background import remove_idist_background
from .psd_moments import psd_moments
from .rotate_tensor import rotate_tensor
from .calculate_epsilon import calculate_epsilon
from .dft_time_shift import dft_time_shift
from .estimate_phase_speed import estimate_phase_speed
from .fft_bandpass import fft_bandpass
from .get_pitch_angle_dist import get_pitch_angle_dist
from .make_model_vdf import make_model_vdf
from .psd_rebin import psd_rebin
from .load_ancillary import load_ancillary
from .vdf_omni import vdf_omni
from .spectr_to_dataset import spectr_to_dataset
from .psd2def import psd2def
from .psd2dpf import psd2dpf
from .dpf2psd import dpf2psd
from .def2psd import def2psd
from .copy_files import copy_files
from .vdf_to_e64 import vdf_to_e64
from .dsl2gse import dsl2gse
from .dsl2gsm import dsl2gsm
from .vdf_projection import vdf_projection
from .vdf_elim import vdf_elim
from .get_data import get_data
from .get_variable import get_variable
from .db_get_variable import db_get_variable
from .make_model_kappa import make_model_kappa

# Hot Plasma Composition Analyser (HPCA)
from .get_hpca_dist import get_hpca_dist
from .hpca_calc_anodes import hpca_calc_anodes
from .hpca_energies import hpca_energies
from .hpca_spin_sum import hpca_spin_sum
from .hpca_pad import hpca_pad

# Fly’s Eye Energetic Particle Spectrometer (FEEPS)
from .feeps_active_eyes import feeps_active_eyes
from .feeps_correct_energies import feeps_correct_energies
from .feeps_energy_table import feeps_energy_table
from .feeps_flat_field_corrections import feeps_flat_field_corrections
from .feeps_omni import feeps_omni
from .feeps_pad import feeps_pad
from .feeps_pad_spinavg import feeps_pad_spinavg
from .feeps_pitch_angles import feeps_pitch_angles
from .feeps_remove_bad_data import feeps_remove_bad_data
from .feeps_remove_sun import feeps_remove_sun
from .feeps_sector_spec import feeps_sector_spec
from .feeps_spin_avg import feeps_spin_avg
from .feeps_split_integral_ch import feeps_split_integral_ch
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv
from .get_feeps_alleyes import get_feeps_alleyes

# Energetic Ion Spectrometer (EIS)
from .eis_ang_ang import eis_ang_ang
from .eis_combine_proton_pad import eis_combine_proton_pad
from .eis_combine_proton_spec import eis_combine_proton_spec
# from .eis_correlation import eis_correlation
from .eis_moments import eis_moments
from .eis_omni import eis_omni
from .eis_pad import eis_pad
from .eis_pad_combine_sc import eis_pad_combine_sc
from .eis_pad_spinavg import eis_pad_spinavg
from .eis_proton_correction import eis_proton_correction
from .eis_skymap import eis_skymap
from .eis_skymap_combine_sc import eis_skymap_combine_sc
from .eis_spec_combine_sc import eis_spec_combine_sc
from .eis_spin_avg import eis_spin_avg
from .get_eis_allt import get_eis_allt
from .eis_combine_proton_skymap import eis_combine_proton_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def mva(inp, flag: str = "mvar"):
    r"""Compute the minimum variance frame.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the quantity to find minimum variance frame.
    flag : {"mvar", "<bn>=0", "td"}, Optional
        Constrain. Default is "mvar".

    Returns
    -------
    out : xarray.DataArray
        Time series of the input quantity in LMN coordinates.
    l : ndarray
        Eigenvalues l[0] > l[1] > l[2].
    lmn : ndarray
        Eigenvectors LMN coordinates.

    See also
    --------
    pyrfu.pyrf.new_xyz

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)

    Compute MVA frame

    >>> b_lmn, lamb, frame = pyrf.mva(b_xyz)

    """

    inp_data = inp.data
    n_t = inp_data.shape[0]

    idx_1, idx_2 = [[0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]]

    if flag in ["mvar", "<bn>=0"]:
        m_mu_nu_m = np.mean(inp_data[:, idx_1] * inp_data[:, idx_2], 0)
        m_mu_nu_m -= np.mean(inp_data, 0)[idx_1] * np.mean(inp_data, 0)[idx_2]

    elif flag.lower() == "td":
        m_mu_nu_m = np.mean(inp_data[:, idx_1] * inp_data[:, idx_2], 0)

    else:
        raise ValueError("invalid flag")

    m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]],
                        m_mu_nu_m[[3, 1, 5]],
                        m_mu_nu_m[[4, 5, 2]]])

    # Compute eigenvalues and eigenvectors
    [l, lmn] = np.linalg.eig(m_mu_nu)

    # Sort eigenvalues
    l, lmn = [l[l.argsort()[::-1]], lmn[:, l.argsort()[::-1]]]

    # ensure that the frame is right handed
    lmn[:, 2] = np.cross(lmn[:, 0], lmn[:, 1])

    if flag.lower() == "<bn>=0":
        inp_mvar_mean = np.mean(np.sum(
            np.tile(inp_data, (3, 1, 1)) * np.transpose(
                np.tile(lmn, (n_t, 1, 1)), (2, 0, 1)), 1), 1)

        a = np.sum(inp_mvar_mean ** 2)

        b = -(l[1] + l[2]) * inp_mvar_mean[0] ** 2
        b -= (l[0] + l[2]) * inp_mvar_mean[1] ** 2
        b -= (l[0] + l[1]) * inp_mvar_mean[2] ** 2

        c = l[1] * l[2] * inp_mvar_mean[0] ** 2
        c += l[0] * l[2] * inp_mvar_mean[1] ** 2
        c += l[0] * l[1] * inp_mvar_mean[2] ** 2

        r = np.roots([a, b, c])

        l_min = np.min(r)

        n = inp_mvar_mean / (l - l_min)
        n /= np.linalg.norm(n, keepdims=True)
        n = np.matmul(lmn, n)

        bn = np.sum(inp_data * np.tile(n, (n_t, 1)), axis=1)

        inp_data_2 = inp_data.copy()
        inp_data_2 -= np.tile(bn, (3, 1)).T * np.tile(n, (n_t, 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, idx_1] * inp_data_2[:, idx_2], 0)
        m_mu_nu_m -= inp_data_2_m[idx_1] * inp_data_2_m[idx_2]

        m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]],
                            m_mu_nu_m[[3, 1, 5]],
                            m_mu_nu_m[[4, 5, 2]]])

        l, lmn = np.linalg.eig(m_mu_nu)

        l, lmn = [l[l.argsort()[::-1]], lmn[:, l.argsort()[::-1]]]

        l[2], lmn[:, 2] = [l_min, np.cross(lmn[:, 0], lmn[:, 1])]

    elif flag.lower() == "td":
        ln = l[2]
        bn = np.sum(inp_data * np.tile(lmn[:, 2], (n_t, 1)), axis=1)

        inp_data_2 = inp_data.copy()
        inp_data_2 -= np.tile(bn, (3, 1)).T * np.tile(lmn[:, 2], (n_t, 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, idx_1] * inp_data_2[:, idx_2], 0)
        m_mu_nu_m -= inp_data_2_m[idx_1] * inp_data_2_m[idx_2]

        m_mu_nu = np.array([m_mu_nu_m[[0, 3, 4]],
                            m_mu_nu_m[[3, 1, 5]],
                            m_mu_nu_m[[4, 5, 2]]])

        l, lmn = np.linalg.eig(m_mu_nu)

        l, lmn = [l[l.argsort()[::-1]], lmn[:, l.argsort()[::-1]]]

        l[2], lmn[:, 2] = [ln, np.cross(lmn[:, 0], lmn[:, 1])]

    out_data = (lmn.T @ inp_data.T).T

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)

    return out, l, lmn



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2021"
__license__ = "MIT"
__version__ = "2.3.7"
__status__ = "Prototype"


def calc_ag(p_xyz):
    r"""Computes agyrotropy coefficient as in [16]_

    .. math::

        AG^{1/3} = \\frac{|\\operatorname[det]{\\mathbf{P}}
        - \\operatorname[det]{\\mathbf{P}}|}
        {\\operatorname[det]{\\mathbf{P}}
        + \\operatorname[det]{\\mathbf{P}}}


    Parameters
    ----------
    p_xyz : xarray.DataArray
        Time series of the pressure tensor

    Returns
    -------
    agyrotropy : xarray.DataArray
        Time series of the agyrotropy coefficient of the specie.

    References
    ----------
    .. [16] H. Che, C. Schiff, G. Le, J. C. Dorelli, B. L. Giles, and T.
            E. Moore (2018), Quantifying the effect of non-Larmor motion
            of electrons on the pres- sure tensor, Phys. Plasmas 25(3),
            032101, doi: https://doi.org/10.1063/1.5016853.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> ic = 1

    Load magnetic field and electron pressure tensor

    >>> b_xyz = mms.get_data("b_gse_fgm_srvy_l2", tint, 1)
    >>> p_xyz_e = mms.get_data("pe_gse_fpi_fast_l2", tint, 1)

    Rotate electron pressure tensor to field aligned coordinates

    >>> p_fac_e_pp = mms.rotate_tensor(p_xyz_e, "fac", b_xyz, "pp")

    Compute agyrotropy coefficient

    >>> ag_e, ag_cr_e = pyrf.calc_ag(p_fac_e_pp)

    """

    # Diagonal and off-diagonal terms
    p_11, p_22, _ = [p_xyz[:, 0, 0], p_xyz[:, 1, 1], p_xyz[:, 2, 2]]
    p_12, p_13, p_23 = [p_xyz[:, 0, 1], p_xyz[:, 0, 2], p_xyz[:, 1, 2]]

    det_p = p_11 * (p_22 ** 2 - p_23 ** 2)
    det_p -= p_12 * (p_12 * p_22 - p_23 * p_13)
    det_p += p_13 * (p_12 * p_23 - p_22 * p_13)

    det_g = p_11 * p_22 ** 2

    agyrotropy = np.abs(det_p - det_g) / (det_p + det_g)

    return agyrotropy
